import os
import shutil
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from omegaconf import DictConfig
import pandas as pd

from autrainer.transforms import SmartCompose

from .abstract_dataset import BaseClassificationDataset
from .utils import ZipDownloadManager


FILES = {
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.1.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.1.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.2.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.2.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.3.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.3.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.4.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.4.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.5.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.5.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.6.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.6.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.7.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.7.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.8.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.8.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.9.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.9.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.10.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.10.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.11.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.11.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.12.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.12.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.13.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.13.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.14.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.14.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.15.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.15.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.audio.16.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.audio.16.zip?download=1",
    "TAU-urban-acoustic-scenes-2020-mobile-development.meta.zip": "https://zenodo.org/records/3819968/files/TAU-urban-acoustic-scenes-2020-mobile-development.meta.zip?download=1",
}

SCENE_CATEGORIES = {
    "indoor": ["airport", "shopping_mall", "metro_station"],
    "outdoor": [
        "park",
        "public_square",
        "street_pedestrian",
        "street_traffic",
    ],
    "transportation": ["bus", "metro", "tram"],
}


class DCASE2020Task1A(BaseClassificationDataset):
    def __init__(
        self,
        path: str,
        features_subdir: str,
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: Union[str, List[str]],
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        batch_size: int,
        inference_batch_size: Optional[int] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
        dev_split: float = 0.0,
        dev_split_seed: Optional[int] = None,
        scene_category: Optional[str] = None,
        exclude_cities: Optional[List[str]] = None,
    ) -> None:
        """TAU Urban Acoustic Scenes 2020 Mobile Task 1 Subtask A
        (DCASE2020Task1A) dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            batch_size: Batch size.
            inference_batch_size: Inference batch size. If None, defaults to
                batch_size. Defaults to None.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
            dev_split: Fraction of the training set to use as the development
                set. Defaults to 0.0.
            dev_split_seed: Seed for the development split. If None, seed is
                used. Defaults to None.
            scene_category: Scene category in ["indoor", "outdoor",
                "transportation"]. Defaults to None.
            exclude_cities: List of cities to exclude from the dataset.
                Defaults to None.
        """
        self._assert_dev_split(dev_split)
        self.dev_split = dev_split
        self.dev_split_seed = dev_split_seed or seed
        self._assert_scene_category(scene_category)
        self.scene_category = scene_category
        self.exclude_cities = exclude_cities
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            batch_size=batch_size,
            inference_batch_size=inference_batch_size,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )

    @staticmethod
    def _assert_scene_category(scene_category: Optional[str]) -> None:
        if scene_category is None:
            return
        if scene_category not in SCENE_CATEGORIES.keys():
            raise ValueError(
                f"Scene category '{scene_category}' must be one of "
                f"{list(SCENE_CATEGORIES.keys())}."
            )

    def load_dataframes(
        self,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        df_train = pd.read_csv(os.path.join(self.path, "train.csv"))
        df_dev = pd.read_csv(os.path.join(self.path, "test.csv"))
        df_test = pd.read_csv(os.path.join(self.path, "test.csv"))

        if self.dev_split > 0:
            df_train, df_dev = self._split_train_dataset(df_train)

        df_train = self._filter_df_by_category(df_train, self.scene_category)
        df_dev = self._filter_df_by_category(df_dev, self.scene_category)
        df_test = self._filter_df_by_category(df_test, self.scene_category)

        if self.exclude_cities is not None:
            df_train = df_train.loc[
                ~df_train["city"].isin(self.exclude_cities)
            ]

        return df_train, df_dev, df_test

    def _split_train_dataset(
        self,
        df: pd.DataFrame,
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Split the training dataset into training and development sets
        taking into account the location.
        Data is split to include at least the `dev_split` fraction of the
        training data in the development set, e.g. favoring the development
        set and the split is not exactly `dev_split`.


        Args:
            df (pd.DataFrame): Training dataset.

        Returns:
            Tuple[pd.DataFrame, pd.DataFrame]: Training and development
            datasets.
        """
        self.rng = np.random.default_rng(self.dev_split_seed)
        unique_locations = df["location"].unique()
        self.rng.shuffle(unique_locations)
        train_size = len(df)
        train_locations = []
        dev_locations = []
        dev_size = 0

        for loc in unique_locations:
            if dev_size < train_size * self.dev_split:
                dev_locations.append(loc)
                dev_size += len(df[df["location"] == loc])
            else:
                train_locations.append(loc)

        return (
            df[df["location"].isin(train_locations)].copy(),
            df[df["location"].isin(dev_locations)].copy(),
        )

    def _filter_df_by_category(
        self,
        df: pd.DataFrame,
        category: Optional[str],
    ) -> pd.DataFrame:
        if category is None:
            return df
        return df.loc[df[self.target_column].isin(SCENE_CATEGORIES[category])]

    @staticmethod
    def download(path: str) -> None:  # pragma: no cover
        """Download the TAU Urban Acoustic Scenes 2020 Mobile Task 1 Subtask A
        (DCASE2020Task1A) dataset.

        As no labels are provided for the evaluation set, the provided training
        and test split of the development set is created.
        Therefore, this download does not include the evaluation set.

        For more information on the dataset, see:
        https://dcase.community/challenge2020/task-acoustic-scene-classification

        Args:
            path: Path to the directory to download the dataset to.
        """

        def _extract_metadata(df: pd.DataFrame) -> pd.DataFrame:
            df["city"] = df["filename"].apply(lambda x: x.split("-")[1])
            df["location"] = df["filename"].apply(lambda x: x.split("-")[2])
            df["segment"] = df["filename"].apply(lambda x: x.split("-")[3])
            df["device"] = df["filename"].apply(
                lambda x: x.split("-")[4].split(".")[0]
            )
            return df

        out_path = os.path.join(path, "default")
        if os.path.isdir(out_path):
            return
        os.makedirs(out_path, exist_ok=True)

        # download and extract files
        dl_manager = ZipDownloadManager(FILES, path)
        dl_manager.download(
            check_exist=["TAU-urban-acoustic-scenes-2020-mobile-development"]
        )
        dl_manager.extract(
            check_exist=["TAU-urban-acoustic-scenes-2020-mobile-development"]
        )
        dev_path = os.path.join(
            path, "TAU-urban-acoustic-scenes-2020-mobile-development"
        )

        # move audio files
        shutil.move(
            os.path.join(dev_path, "audio"),
            os.path.join(out_path),
        )

        # load, extract, and save dataframes
        df_train = pd.read_csv(
            os.path.join(dev_path, "evaluation_setup", "fold1_train.csv"),
            sep="\t",
        )
        df_test = pd.read_csv(
            os.path.join(dev_path, "evaluation_setup", "fold1_evaluate.csv"),
            sep="\t",
        )
        df_train = _extract_metadata(df_train)
        df_test = _extract_metadata(df_test)
        df_train.to_csv(os.path.join(path, "train.csv"), index=False)
        df_test.to_csv(os.path.join(path, "test.csv"), index=False)

        # remove unnecessary files
        shutil.rmtree(dev_path)



import os
from typing import Callable, List
import warnings

import requests
import torch
import torch.nn.functional as F


def _download_weights(url: str, destination: str) -> None:
    print(f"Downloading weights from {url}")
    try:
        req = requests.get(url)
        if req.status_code == 200:
            with open(destination, "wb") as f:
                f.write(req.content)
            print(f"Placed weights in {destination}")
        else:
            req.raise_for_status()
    except Exception as e:
        raise ValueError(
            f"Failed to download weights from {url} due to {e}."
        ) from e


def load_transfer_weights(model: torch.nn.Module, weights_link: str) -> None:
    weights_name = os.path.basename(weights_link)
    hub_dir = os.path.join(torch.hub.get_dir(), "checkpoints")
    weights_path = os.path.join(hub_dir, weights_name)
    if not os.path.exists(weights_path):
        os.makedirs(os.path.dirname(weights_path), exist_ok=True)
        _download_weights(weights_link, weights_path)
    if not os.path.exists(weights_path):
        raise ValueError(f"Failed to download weights to {weights_path}.")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        model.load_state_dict(
            torch.load(
                weights_path,
                map_location="cpu",
                weights_only=True,
            )["model"],
            strict=False,
        )


def init_layer(layer):
    """Initialize a Linear or Convolutional layer."""
    torch.nn.init.xavier_uniform_(layer.weight)

    if hasattr(layer, "bias"):
        if layer.bias is not None:
            layer.bias.data.fill_(0.0)


def init_bn(bn):
    """Initialize a Batchnorm layer."""
    bn.bias.data.fill_(0.0)
    bn.weight.data.fill_(1.0)


class ConvBlock(torch.nn.Module):
    def __init__(self, in_channels: int, out_channels: int):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels

        self.conv1 = torch.nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=(3, 3),
            stride=(1, 1),
            padding=(1, 1),
            bias=False,
        )

        self.conv2 = torch.nn.Conv2d(
            in_channels=self.out_channels,
            out_channels=self.out_channels,
            kernel_size=(3, 3),
            stride=(1, 1),
            padding=(1, 1),
            bias=False,
        )

        self.bn1 = torch.nn.BatchNorm2d(self.out_channels)
        self.bn2 = torch.nn.BatchNorm2d(self.out_channels)

        self.init_weight()

    def init_weight(self):
        init_layer(self.conv1)
        init_layer(self.conv2)
        init_bn(self.bn1)
        init_bn(self.bn2)

    def forward(self, x, pool_size=(2, 2), pool_type="avg"):
        x = F.relu_(self.bn1(self.conv1(x)))
        x = F.relu_(self.bn2(self.conv2(x)))
        if pool_type == "max":
            x = F.max_pool2d(x, kernel_size=pool_size)
        elif pool_type == "avg":
            x = F.avg_pool2d(x, kernel_size=pool_size)
        elif pool_type == "avg+max":
            x1 = F.avg_pool2d(x, kernel_size=pool_size)
            x2 = F.max_pool2d(x, kernel_size=pool_size)
            x = x1 + x2
        else:
            raise Exception("Incorrect argument!")

        return x


class ExtractLayerEmbeddings:
    def __init__(
        self,
        model: torch.nn.Module,
        layer_selector_fn: Callable = None,
    ) -> None:
        """Extract embeddings from a layer in a model.

        Args:
            model: Model to extract embeddings from.
            layer_selector_fn: Function to select the layer to extract
                embeddings from. The model is passed to this function and
                the output should be the layer. If None, the preceding layer
                of the last linear layer is selected. Defaults to None.
        """
        self.model = model
        self._embeddings = None
        self._hook_handle = None
        self._hook_layer = (
            layer_selector_fn(model)
            if layer_selector_fn is not None
            else self._find_second_to_last_layer(model)
        )

    @staticmethod
    def _find_second_to_last_layer(model: torch.nn.Module) -> torch.nn.Module:
        def _flatten_layers(
            layer: torch.nn.Module,
            layers: List[torch.nn.Module],
        ) -> None:
            if len(list(layer.children())) == 0:
                layers.append(layer)
            else:
                for child in layer.children():
                    _flatten_layers(child, layers)

        layers = []
        _flatten_layers(model, layers)
        if len(layers) < 2:
            raise ValueError("Model must have at least two layers.")
        idx = max(
            idx
            for idx, layer in enumerate(layers)
            if isinstance(layer, torch.nn.Linear)
        )
        return layers[idx - 1]

    def _hook(
        self,
        module: torch.nn.Module,
        input: torch.Tensor,
        output: torch.Tensor,
    ) -> None:
        self._embeddings = output

    def _register(self) -> None:
        if self._hook_handle is not None:
            return
        self._hook_handle = self._hook_layer.register_forward_hook(self._hook)

    def _unregister(self) -> None:
        if self._hook_handle is None:
            return
        self._hook_handle.remove()
        self._hook_handle = None

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return self.embeddings(x)

    def embeddings(self, x: torch.Tensor) -> torch.Tensor:
        self._register()
        self.model(x)
        self._unregister()
        return self._embeddings



import os
import shutil
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from omegaconf import DictConfig
import pandas as pd

from autrainer.transforms import SmartCompose

from .abstract_dataset import BaseClassificationDataset
from .utils import ZipDownloadManager


FILES = {
    "ff1010bird_wav.zip": "https://archive.org/download/ff1010bird/ff1010bird_wav.zip",
    "ff1010bird_metadata_2018.csv": "https://ndownloader.figshare.com/files/10853303",
    "BirdVox-DCASE-20k.zip": "https://zenodo.org/record/1208080/files/BirdVox-DCASE-20k.zip",
    "BirdVoxDCASE20k_csvpublic.csv": "https://ndownloader.figshare.com/files/10853300",
    "warblrb10k_public_wav.zip": "https://archive.org/download/warblrb10k_public/warblrb10k_public_wav.zip",
    "warblrb10k_public_metadata_2018.csv": "https://ndownloader.figshare.com/files/10853306",
}


class DCASE2018Task3(BaseClassificationDataset):
    def __init__(
        self,
        path: str,
        features_subdir: str,
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: Union[str, List[str]],
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        batch_size: int,
        inference_batch_size: Optional[int] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
        dev_split: float = 0.0,
        dev_split_seed: Optional[int] = None,
    ) -> None:
        """DCASE 2018 Task 3 dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            batch_size: Batch size.
            inference_batch_size: Inference batch size. If None, defaults to
                batch_size. Defaults to None.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
            dev_split: Fraction of the training set to use as the development
                set. Defaults to 0.0.
            dev_split_seed: Seed for the development split. If None, seed is
                used. Defaults to None.
        """
        self._assert_dev_split(dev_split)
        self.dev_split = dev_split
        self.dev_split_seed = dev_split_seed or seed
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            batch_size=batch_size,
            inference_batch_size=inference_batch_size,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )

    def load_dataframes(
        self,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        df_train = pd.read_csv(os.path.join(self.path, "train.csv"))
        df_dev = pd.read_csv(os.path.join(self.path, "test.csv"))
        df_test = pd.read_csv(os.path.join(self.path, "test.csv"))

        if self.dev_split > 0:
            df_train, df_dev = self._split_train_dataset(df_train)

        return df_train, df_dev, df_test

    def _split_train_dataset(
        self,
        df: pd.DataFrame,
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        rng = np.random.default_rng(self.dev_split_seed)
        indices = rng.permutation(df.index)
        train_indices = indices[: int(len(indices) * (1 - self.dev_split))]
        return (
            df[df.index.isin(train_indices)].copy(),
            df[~df.index.isin(train_indices)].copy(),
        )

    @staticmethod
    def download(path: str) -> None:  # pragma: no cover
        """Download the DCASE 2018 Task 3 dataset.

        For the train dataset, the following subsets are used:

        - Field recordings, worldwide ("freefield1010")
        - Remote monitoring flight calls, USA ("BirdVox-DCASE-20k")

        For the test dataset, the following subset is used:

        - Crowdsourced dataset, UK ("warblrb10k")

        Both the training and test datasets are taken from the development set,
        as no labels are provided for the evaluation set.
        For more information on the dataset, see:
        https://dcase.community/challenge2018/task-bird-audio-detection

        Args:
            path: Path to the directory to download the dataset to.
        """

        out_path = os.path.join(path, "default")
        if os.path.isdir(out_path):
            return

        # download and extract files
        dl_manager = ZipDownloadManager(FILES, path)
        dl_manager.download(check_exist=["wav", "warblrb10k_public_wav"])
        dl_manager.extract(check_exist=["wav", "warblrb10k_public_wav"])

        # move audio files
        shutil.move(os.path.join(path, "wav"), out_path)

        # process dataframes
        ff1010_path = os.path.join(path, "ff1010bird_metadata_2018.csv")
        birdvox_path = os.path.join(path, "BirdVoxDCASE20k_csvpublic.csv")
        warblrb_path = os.path.join(
            path, "warblrb10k_public_metadata_2018.csv"
        )
        df_ff1010 = pd.read_csv(ff1010_path).rename(
            columns={"itemid": "filename", "datasetid": "dataset"}
        )
        df_birdvox = pd.read_csv(birdvox_path).rename(
            columns={"itemid": "filename", "datasetid": "dataset"}
        )
        df_warblrb10k = pd.read_csv(warblrb_path).rename(
            columns={"itemid": "filename", "datasetid": "dataset"}
        )
        df_train = pd.concat([df_ff1010, df_birdvox])
        df_test = df_warblrb10k
        df_train["filename"] = df_train["filename"].apply(lambda x: f"{x}.wav")
        df_test["filename"] = df_test["filename"].apply(lambda x: f"{x}.wav")
        df_train.to_csv(os.path.join(path, "train.csv"), index=False)
        df_test.to_csv(os.path.join(path, "test.csv"), index=False)

        # remove unnecessary files
        os.remove(ff1010_path)
        os.remove(birdvox_path)
        os.remove(warblrb_path)



import torch

from .abstract_model import AbstractModel
from .ffnn import FFNN


class Sequential(AbstractModel):
    def __init__(
        self,
        input_dim: int,
        hidden_size: int,
        num_layers: int,
        dropout: float = 0.5,
        cell: str = "LSTM",
        time_pooling: bool = True,
        bidirectional: bool = False,
    ):
        super().__init__(
            output_dim=2 * hidden_size if bidirectional else hidden_size
        )
        self.input_dim = input_dim
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.dropout = dropout
        self.time_pooling = time_pooling
        self.bidirectional = bidirectional
        self.cell = cell
        if self.cell not in ["LSTM", "GRU"]:
            raise NotImplementedError(self.cell)
        cell_cls = torch.nn.LSTM if self.cell == "LSTM" else torch.nn.GRU
        self.model = cell_cls(
            input_size=input_dim,
            hidden_size=hidden_size,
            batch_first=True,
            num_layers=num_layers,
            dropout=dropout,
            bidirectional=bidirectional,
        )

    def embeddings(self, x: torch.Tensor) -> torch.Tensor:
        x = self.model(x)[0]
        if self.time_pooling:
            x = x.mean(1)
        return x

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.embeddings(x)


class SeqFFNN(AbstractModel):
    def __init__(
        self,
        output_dim: int,
        backbone_input_dim: int,
        backbone_hidden_size: int,
        backbone_num_layers: int,
        hidden_size: int,
        num_layers: int = 2,
        sigmoid: bool = False,
        softmax: bool = False,
        dropout: float = 0.5,
        backbone_dropout: float = 0.5,
        backbone_cell: str = "LSTM",
        backbone_time_pooling: bool = True,
        backbone_bidirectional: bool = False,
    ):
        """Sequential model with FFNN frontend.

        Args:
            output_dim: Output dimension of the FFNN.
            backbone_input_dim: Input dimension of the backbone.
            backbone_hidden_size: Hidden size of the backbone.
            backbone_num_layers: Number of layers of the backbone.
            hidden_size: Hidden size of the FFNN.
            num_layers: Number of layers of the FFNN. Defaults to 2.
            sigmoid: Whether to apply sigmoid activation. Defaults to False.
            softmax: Whether to apply softmax activation. Defaults to False.
            dropout: Dropout rate of the FFNN. Defaults to 0.5.
            backbone_dropout: Dropout rate of the backbone. Defaults to 0.5.
            backbone_cell: Cell type of the backbone in ["LSTM", "GRU"].
                Defaults to "LSTM".
            backbone_time_pooling: Whether to apply time pooling in the
                backbone. Defaults to True.
            backbone_bidirectional: Whether to use a bidirectional backbone.
                Defaults to False.
        """
        super().__init__(output_dim)
        self.backbone_input_dim = backbone_input_dim
        self.backbone_hidden_size = backbone_hidden_size
        self.backbone_num_layers = backbone_num_layers
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.sigmoid = sigmoid
        self.softmax = softmax
        self.dropout = dropout
        self.backbone_dropout = backbone_dropout
        self.backbone_cell = backbone_cell
        self.backbone_time_pooling = backbone_time_pooling
        self.backbone_bidirectional = backbone_bidirectional
        self.backbone = Sequential(
            input_dim=backbone_input_dim,
            hidden_size=backbone_hidden_size,
            num_layers=backbone_num_layers,
            dropout=backbone_dropout,
            cell=backbone_cell,
            time_pooling=backbone_time_pooling,
            bidirectional=backbone_bidirectional,
        )
        self.frontend = FFNN(
            input_size=self.backbone.output_dim,
            hidden_size=hidden_size,
            output_dim=output_dim,
            num_layers=num_layers,
            sigmoid=sigmoid,
            softmax=softmax,
            dropout=dropout,
        )

    def embeddings(self, x: torch.Tensor) -> torch.Tensor:
        return self.backbone(x.squeeze(1))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.frontend(self.embeddings(x))



from abc import ABC, abstractmethod
from typing import Optional

import audiofile
import audobject
import numpy as np
import torch
import torchaudio
import torchvision


class AbstractFileHandler(ABC, audobject.Object):
    def __init__(self) -> None:
        """Abstract file handler for loading files in the dataset and saving
        files during preprocessing.

        Serves as the base for creating custom file handlers that handle
        loading and saving of different file types.
        """

    def __call__(self, file: str) -> torch.Tensor:
        """Load a file from a path.

        Args:
            file: Path to file.

        Returns:
            Loaded file.
        """
        return self.load(file)

    @abstractmethod
    def load(self, file: str) -> torch.Tensor:
        """Load a file from a path.

        Args:
            file: Path to file.

        Returns:
            Loaded file.
        """

    @abstractmethod
    def save(self, file: str, data: torch.Tensor) -> None:
        """Save a file to a path.

        Args:
            file: Path to file.
            data: Data to save.
        """


class IdentityFileHandler(AbstractFileHandler):
    def __init__(self) -> None:
        """Identity file handler serving as a no-op. Both load and save methods
        return None.
        """

    def load(self, file: str) -> None:
        """Identity operation.

        Args:
            file: Path to file.
        """
        return file

    def save(self, file: str, data: torch.Tensor) -> None:
        """Identity operation.

        Args:
            file: Path to file.
            data: Data to save.
        """


class ImageFileHandler(AbstractFileHandler):
    def __init__(self) -> None:
        """Image file handler for loading and saving with torchvision.
        Torchvision supports the PNG, JPEG, and GIF image formats for loading
        and saving images.
        """

    def load(self, file: str) -> torch.Tensor:
        """Load an image from a file as a uint8 tensor in the range [0, 255].

        Args:
            file: Path to image file.

        Returns:
            Uint8 image tensor.
        """
        return torchvision.io.read_image(file)

    def save(self, file: str, data: torch.Tensor) -> None:
        """Save an image tensor to a file.

        If the tensor is of type uint8, it is assumed to be in the range
        [0, 255] and divided by 255 before saving.

        Args:
            file: Path to image file.
            data: Image tensor to save.
        """
        if data.dtype == torch.uint8:
            data = data / 255
        torchvision.utils.save_image(data, file)


class NumpyFileHandler(AbstractFileHandler):
    def __init__(self) -> None:
        """Numpy file handler for loading and saving numpy arrays."""

    def load(self, file: str) -> torch.Tensor:
        """Load a numpy array from a file.

        Args:
            file: Path to numpy file.

        Returns:
            Numpy array as a tensor.
        """
        return torch.from_numpy(np.load(file))

    def save(self, file: str, data: torch.Tensor) -> None:
        """Save a tensor to a numpy file.

        Args:
            file: Path to numpy file.
            data: Tensor to save.
        """
        np.save(file, data.numpy())


class AudioFileHandler(AbstractFileHandler):
    def __init__(
        self,
        target_sample_rate: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Audio file handler with optional resampling.

        Args:
            target_sample_rate: Target sample rate to resample audio files to
                during loading. Has to be specified to save audio files.
                If None, audio files are loaded with their
                original sample rate. Defaults to None.
            **kwargs: Additional keyword arguments passed to
                torchaudio.transforms.Resample.
        """
        # ? audobject passes _object_root_ to the object, which is not a valid
        # ? argument for the file handler.
        kwargs.pop("_object_root_", None)
        self.target_sample_rate = target_sample_rate
        self.kwargs = kwargs

    def load(self, file: str) -> torch.Tensor:
        """Load an audio file and resample it if a target sample rate is
        specified.

        Args:
            file: Path to audio file.

        Returns:
            Loaded audio file as a tensor.
        """
        x, sr = audiofile.read(file, always_2d=True)
        if (
            self.target_sample_rate is not None
            and sr != self.target_sample_rate
        ):
            resample = torchaudio.transforms.Resample(
                sr,
                self.target_sample_rate,
                **self.kwargs,
            )
            x = resample(torch.from_numpy(x)).numpy()
        return torch.from_numpy(x)

    def save(self, file: str, data: torch.Tensor) -> None:
        """Save an audio tensor to a file.

        Args:
            file: Path to audio file.
            data: Audio data to save.

        Raises:
            ValueError: If target sample rate is not specified.
        """
        if self.target_sample_rate is None:
            raise ValueError(
                "Target sample rate has to be specified to save audio files."
            )
        audiofile.write(file, data.numpy(), self.target_sample_rate)



from abc import ABC, abstractmethod
import functools
import logging
from typing import Callable, Union
import warnings

import numpy as np
import pandas as pd


def ignore_runtime_warning(func):
    @functools.wraps(func)
    def wrapper_ignore_warning(*args, **kwargs):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            return func(*args, **kwargs)

    return wrapper_ignore_warning


class AbstractMetric(ABC):
    def __init__(
        self,
        name: str,
        fn: Callable,
        fallback: float,
        **fn_kwargs: dict,
    ) -> None:
        """Abstract class for metrics.

        Args:
            name: The name of the metric.
            fn: The function to compute the metric.
            fallback: The fallback value if the metric is NaN.
            **fn_kwargs: Additional keyword arguments to pass to the function.
        """
        self.name = name
        self.fn = fn
        self.fallback = fallback
        self.fn_kwargs = fn_kwargs
        self.logger = logging.getLogger(__name__)

    @ignore_runtime_warning
    def __call__(self, *args, **kwargs) -> float:
        """Compute the metric.

        Args:
            *args: Positional arguments for the function.
            **kwargs: Keyword arguments for the function.

        Returns:
            The score.
        """
        try:
            score = self.fn(*args, **kwargs, **self.fn_kwargs)
        except ValueError as e:
            self.logger.warning(
                f"Error computing {self.name} metric due to: {e}. "
                f"Defaulting to fallback value of {self.fallback}."
            )
            return self.fallback
        if np.isnan(score):
            self.logger.warning(
                f"Error computing {self.name} metric as score is NaN. "
                f"Defaulting to fallback value of {self.fallback}."
            )
            return self.fallback
        return float(score)

    @property
    @abstractmethod
    def starting_metric(self) -> float:
        """The starting metric value.

        Returns:
            The starting metric value.
        """

    @property
    @abstractmethod
    def suffix(self) -> str:
        """The suffix of the metric.

        Returns:
            The suffix of the metric.
        """

    @staticmethod
    @abstractmethod
    def get_best(a: Union[pd.Series, np.ndarray]) -> float:
        """Get the best metric value from a series of scores.

        Args:
            a: Pandas series or numpy array of scores.

        Returns:
            Best metric value.
        """

    @staticmethod
    @abstractmethod
    def get_best_pos(a: Union[pd.Series, np.ndarray]) -> int:
        """Get the position of the best metric value from a series of scores.

        Args:
            a: Pandas series or numpy array of scores.

        Returns:
            Position of the best metric value.
        """

    @staticmethod
    @abstractmethod
    def compare(a: Union[int, float], b: Union[int, float]) -> bool:
        """Compare two scores and return True if the first score is better.

        Args:
            a: First score.
            b: Second score.

        Returns:
            True if the first score is better.
        """


class BaseAscendingMetric(AbstractMetric):
    def __init__(
        self,
        name: str,
        fn: Callable,
        fallback: float = None,
        **fn_kwargs: dict,
    ) -> None:
        """Base for ascending metrics with higher values being better.

        Args:
            name: The name of the metric.
            fn: The function to compute the metric.
            fallback: The fallback value if the metric is NaN. If None,
                the fallback value is set to -1e32. Defaults to None.
            **fn_kwargs: Additional keyword arguments to pass to the function.
        """
        super().__init__(name, fn, fallback or -1e32, **fn_kwargs)

    @property
    def starting_metric(self) -> float:
        """Ascending metric starting value.

        Returns:
            -1e32
        """
        return -1e32

    @property
    def suffix(self) -> str:
        """Ascending metric suffix.

        Returns:
            "max"
        """
        return "max"

    @staticmethod
    def get_best(a: Union[pd.Series, np.ndarray]) -> float:
        return float(a.max())

    @staticmethod
    def get_best_pos(a: Union[pd.Series, np.ndarray]) -> int:
        if isinstance(a, pd.Series):
            return int(a.idxmax())
        return int(a.argmax())

    @staticmethod
    def compare(a: Union[int, float], b: Union[int, float]) -> bool:
        return a > b


class BaseDescendingMetric(AbstractMetric):
    def __init__(
        self,
        name: str,
        fn: Callable,
        fallback: float = None,
        **fn_kwargs: dict,
    ) -> None:
        """Base for descending metrics with lower values being better.

        Args:
            name: The name of the metric.
            fn: The function to compute the metric.
            fallback: The fallback value if the metric is NaN. If None,
                the fallback value is set to 1e32. Defaults to None.
            **fn_kwargs: Additional keyword arguments to pass to the function.
        """
        super().__init__(name, fn, fallback or 1e32, **fn_kwargs)

    @property
    def starting_metric(self) -> float:
        """Descending metric starting value.

        Returns:
            1e32
        """
        return 1e32

    @property
    def suffix(self) -> str:
        """Descending metric suffix.

        Returns:
            "min"
        """
        return "min"

    @staticmethod
    def get_best(a: Union[pd.Series, np.ndarray]) -> float:
        return float(a.min())

    @staticmethod
    def get_best_pos(a: Union[pd.Series, np.ndarray]) -> int:
        if isinstance(a, pd.Series):
            return int(a.idxmin())
        return int(a.argmin())

    @staticmethod
    def compare(a: Union[int, float], b: Union[int, float]) -> bool:
        return a < b



from .bookkeeping import Bookkeeping
from .hardware import get_hardware_info, save_hardware_info, set_device
from .set_seed import set_seed
from .silence import silence
from .timer import Timer


__all__ = [
    "Bookkeeping",
    "get_hardware_info",
    "save_hardware_info",
    "set_device",
    "set_seed",
    "silence",
    "Timer",
]



from abc import ABC, abstractmethod
from functools import cached_property
import os
from typing import Dict, List, Optional, Tuple, TypeVar, Union

from omegaconf import DictConfig
import pandas as pd
import torch
from torch.utils.data import DataLoader

import autrainer
from autrainer.metrics import AbstractMetric
from autrainer.transforms import SmartCompose

from .utils import (
    AbstractFileHandler,
    AbstractTargetTransform,
    DatasetWrapper,
    LabelEncoder,
    MinMaxScaler,
    MultiLabelEncoder,
)


T = TypeVar("T")
TASKS = ["classification", "regression", "ml-classification"]


class AbstractDataset(ABC):
    def __init__(
        self,
        path: str,
        features_subdir: str,
        seed: int,
        task: str,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: Union[str, List[str]],
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        batch_size: int,
        inference_batch_size: Optional[int] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
    ) -> None:
        """Abstract dataset class.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
            seed: Seed for reproducibility.
            task: Task of the dataset in ['classification', 'regression',
                'ml-classification'].
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            batch_size: Batch size.
            inference_batch_size: Inference batch size. If None, defaults to
                batch_size. Defaults to None.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
        """
        self._assert_task(task)
        self._assert_directory(path, features_subdir)
        self.path = path
        self.features_subdir = features_subdir
        self.seed = seed
        self.task = task
        self.metrics = [self._init_metric(m) for m in metrics]
        self.tracking_metric = self._init_metric(tracking_metric)
        self.index_column = index_column
        self.target_column = target_column
        self.file_type = file_type
        self.file_handler = self._init_file_handler(file_handler)
        self.batch_size = batch_size
        self.inference_batch_size = inference_batch_size or batch_size
        self.train_transform = train_transform or SmartCompose([])
        self.dev_transform = dev_transform or SmartCompose([])
        self.test_transform = test_transform or SmartCompose([])
        self.stratify = stratify or []

        self._generator = torch.Generator().manual_seed(self.seed)
        self.df_train, self.df_dev, self.df_test = self.load_dataframes()
        self._assert_stratify()

    @staticmethod
    def _assert_task(task: str) -> None:
        if task not in TASKS:
            raise ValueError(f"Task '{task}' not supported.")

    @staticmethod
    def _assert_directory(path: str, features_subdir: str) -> None:
        if not os.path.isdir(path):
            raise ValueError(f"Directory path='{path}' does not exist.")
        if not os.path.isdir(os.path.join(path, features_subdir)):
            raise ValueError(
                f"Directory features_subdir='{features_subdir}' "
                f"does not exist in '{path}'."
            )

    def _assert_stratify(self) -> None:
        if self.task == "ml-classification" and self.stratify:
            raise ValueError(
                "Stratify is not supported for 'ml-classification' tasks."
            )

        for column in self.stratify:
            if (
                column not in self.df_dev.columns
                or column not in self.df_test.columns
            ):
                raise ValueError(
                    f"Stratify column '{column}' is not present in "
                    "the development or test dataframes."
                )

    @property
    @abstractmethod
    def target_transform(self) -> AbstractTargetTransform:
        """Get the transform to apply to the target.

        Returns:
            Target transform.
        """

    @property
    @abstractmethod
    def output_dim(self) -> int:
        """Get the output dimension of the dataset.

        Returns:
            Output dimension.
        """

    @staticmethod
    def _init_metric(metric: Union[str, DictConfig, Dict]) -> AbstractMetric:
        return autrainer.instantiate_shorthand(
            config=metric,
            instance_of=AbstractMetric,
        )

    @staticmethod
    def _init_file_handler(
        file_handler: Union[str, DictConfig, Dict],
    ) -> AbstractFileHandler:
        return autrainer.instantiate_shorthand(
            config=file_handler,
            instance_of=AbstractFileHandler,
        )

    def load_dataframes(
        self,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """Load the dataframes.

        Returns:
            Dataframes for training, development, and testing.
        """
        return (
            pd.read_csv(os.path.join(self.path, "train.csv")),
            pd.read_csv(os.path.join(self.path, "dev.csv")),
            pd.read_csv(os.path.join(self.path, "test.csv")),
        )

    def _init_dataset(
        self,
        df: pd.DataFrame,
        transform: SmartCompose,
    ) -> DatasetWrapper:
        """Initialize a wrapper around torch.utils.data.Dataset.

        Args:
            df: Dataframe to use.
            transform: Transform to apply to the features.

        Returns:
            Initialized dataset.
        """
        return DatasetWrapper(
            path=self.path,
            features_subdir=self.features_subdir,
            index_column=self.index_column,
            target_column=self.target_column,
            file_type=self.file_type,
            file_handler=self.file_handler,
            df=df,
            transform=transform,
            target_transform=self.target_transform,
        )

    @cached_property
    def train_dataset(self) -> DatasetWrapper:
        """Get the training dataset.

        Returns:
            Training dataset.
        """
        return self._init_dataset(self.df_train, self.train_transform)

    @cached_property
    def dev_dataset(self) -> DatasetWrapper:
        """Get the development dataset.

        Returns:
            Development dataset.
        """
        return self._init_dataset(self.df_dev, self.dev_transform)

    @cached_property
    def test_dataset(self) -> DatasetWrapper:
        """Get the test dataset.

        Returns:
            Test dataset.
        """
        return self._init_dataset(self.df_test, self.test_transform)

    @cached_property
    def train_loader(self) -> DataLoader:
        """Get the training loader.

        Returns:
            Training loader.
        """
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            generator=self._generator,
            collate_fn=self.train_transform.get_collate_fn(self),
        )

    @cached_property
    def dev_loader(self) -> DataLoader:
        """Get the development loader.

        Returns:
            Development loader.
        """
        return DataLoader(
            self.dev_dataset,
            batch_size=self.inference_batch_size,
            shuffle=False,
            generator=self._generator,
            collate_fn=self.dev_transform.get_collate_fn(self),
        )

    @cached_property
    def test_loader(self) -> DataLoader:
        """Get the test loader.

        Returns:
            Test loader.
        """
        return DataLoader(
            self.test_dataset,
            batch_size=self.inference_batch_size,
            shuffle=False,
            generator=self._generator,
            collate_fn=self.dev_transform.get_collate_fn(self),
        )

    def get_evaluation_data(
        self,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, List[str], AbstractTargetTransform]:
        """Get the evaluation data.

        Returns:
            Dataframes for development and testing, columns to stratify on, and
                the target transform.
        """
        return self.df_dev, self.df_test, self.stratify, self.target_transform

    @staticmethod
    def download(path: str) -> None:
        """Download the dataset. Can be implemented by subclasses, but is not
        required.

        Args:
            path: Path to download the dataset to.
        """


class BaseClassificationDataset(AbstractDataset):
    def __init__(
        self,
        path: str,
        features_subdir: str,
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: Union[str, List[str]],
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        batch_size: int,
        inference_batch_size: Optional[int] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
    ) -> None:
        """Base classification dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            batch_size: Batch size.
            inference_batch_size: Inference batch size. If None, defaults to
                batch_size. Defaults to None.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
        """
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            task="classification",
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            batch_size=batch_size,
            inference_batch_size=inference_batch_size,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )

    @cached_property
    def target_transform(self) -> LabelEncoder:
        return LabelEncoder(
            self.df_train[self.target_column].unique().tolist()
        )

    @cached_property
    def output_dim(self) -> int:
        """Get the output dimension of the dataset.

        Returns:
            Number of classes.
        """
        return len(self.df_train[self.target_column].unique())

    @staticmethod
    def _assert_dev_split(dev_split: float) -> None:
        if not 0 <= dev_split < 1:
            raise ValueError(f"dev_split '{dev_split}' must be in [0, 1).")

    @staticmethod
    def _assert_choice(choice: T, choices: List[T], name: str) -> None:
        if choice not in choices:
            raise ValueError(f"{name} '{choice}' not in {choices}.")


class BaseMLClassificationDataset(AbstractDataset):
    def __init__(
        self,
        path: str,
        features_subdir: str,
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: Union[str, List[str]],
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        batch_size: int,
        inference_batch_size: Optional[int] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
        threshold: float = 0.5,
    ) -> None:
        """Base multi-label classification dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            batch_size: Batch size.
            inference_batch_size: Inference batch size. If None, defaults to
                batch_size. Defaults to None.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
            threshold: Threshold for classification. Defaults to 0.5.
        """
        self._assert_threshold(threshold)
        self.threshold = threshold
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            task="ml-classification",
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            batch_size=batch_size,
            inference_batch_size=inference_batch_size,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )
        self._assert_target_column(allowed_columns=self.df_train.columns)

    @staticmethod
    def _assert_threshold(threshold: float) -> None:
        if not 0 <= threshold <= 1:
            raise ValueError(
                f"Threshold '{threshold}' must be between 0 and 1."
            )

    def _assert_target_column(self, allowed_columns: List[str]) -> None:
        if isinstance(self.target_column, str):
            raise ValueError(
                f"Invalid target column '{self.target_column}'. "
                "Must be a list of strings for multi-label classification."
            )
        if not self.target_column:
            raise ValueError("No target columns provided.")

        if not all(c in allowed_columns for c in self.target_column):
            raise ValueError(
                f"Invalid target columns '{self.target_column}'. "
                f"Must be in {allowed_columns}."
            )

    @cached_property
    def target_transform(self) -> MultiLabelEncoder:
        return MultiLabelEncoder(self.threshold, self.target_column)

    @cached_property
    def output_dim(self) -> int:
        """Get the output dimension of the dataset.

        Returns:
            Number of classes.
        """
        return len(self.target_column)


class BaseRegressionDataset(AbstractDataset):
    def __init__(
        self,
        path: str,
        features_subdir: str,
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: Union[str, List[str]],
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        batch_size: int,
        inference_batch_size: Optional[int] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
    ) -> None:
        """Base regression dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            batch_size: Batch size.
            inference_batch_size: Inference batch size. If None, defaults to
                batch_size. Defaults to None.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
        """
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            task="regression",
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            batch_size=batch_size,
            inference_batch_size=inference_batch_size,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )

    @cached_property
    def target_transform(self) -> MinMaxScaler:
        return MinMaxScaler(
            minimum=self.df_train[self.target_column].min(),
            maximum=self.df_train[self.target_column].max(),
        )

    @cached_property
    def output_dim(self) -> int:
        """Get the output dimension of the dataset.

        Returns:
            Always 1 for regression tasks.
        """
        return 1



from dataclasses import dataclass
import os
import shutil
from typing import List

from autrainer.core.scripts.abstract_script import MockParser

from .abstract_postprocess_script import (
    AbstractPostprocessArgs,
    AbstractPostprocessScript,
)
from .utils import catch_cli_errors


@dataclass
class DeleteFailedArgs(AbstractPostprocessArgs):
    force: bool


class DeleteFailedScript(AbstractPostprocessScript):
    def __init__(self) -> None:
        super().__init__(
            "rm-failed",
            "Delete failed runs from an experiment.",
            epilog="Example: autrainer rm-failed results default",
            dataclass=DeleteFailedArgs,
        )

    def add_arguments(self) -> None:
        super().add_arguments()
        self.parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            default=False,
            help=(
                "Force deletion of failed runs without confirmation. "
                "Defaults to False."
            ),
        )

    def main(self, args: DeleteFailedArgs) -> None:
        super().main(args)
        self._clean_failed(args)

    def _clean_failed(self, args: DeleteFailedArgs) -> None:
        failed_runs = self._collect_failed_runs(args)
        if not failed_runs:
            return
        if args.force or self._confirm_deletion(failed_runs):
            self._delete_runs(args, failed_runs)

    def _collect_failed_runs(self, args: DeleteFailedArgs) -> List[str]:
        path = os.path.join(args.results_dir, args.experiment_id, "training")
        return [
            f
            for f in os.listdir(path)
            if os.path.isdir(os.path.join(path, f))
            and not os.path.exists(os.path.join(path, f, "metrics.csv"))
        ]

    def _confirm_deletion(self, failed_runs: List[str]) -> bool:
        print("The following failed runs will be deleted:")
        print("\n".join(failed_runs))
        print()
        response = input(
            "Are you sure you want to delete "
            f"{len(failed_runs)} failed runs? (y/n): "
        ).lower()
        return response in ["y", "yes"]

    def _delete_runs(
        self,
        args: DeleteFailedArgs,
        failed_runs: List[str],
    ) -> None:
        path = os.path.join(args.results_dir, args.experiment_id, "training")
        for run in failed_runs:
            shutil.rmtree(os.path.join(path, run))


@catch_cli_errors
def rm_failed(
    results_dir: str,
    experiment_id: str,
    force: bool = False,
) -> None:
    """Delete failed runs from an experiment.

    If called in a notebook, the function will not raise an error and print
    the error message instead.

    Args:
        results_dir: Path to grid search results directory.
        experiment_id: ID of experiment to postprocess.
        force: Force deletion of failed runs without confirmation.
            Defaults to False.

    Raises:
        CommandLineError: If the results directory or experiment ID dont exist.
    """
    script = DeleteFailedScript()
    script.parser = MockParser()
    script.main(
        DeleteFailedArgs(
            results_dir,
            experiment_id,
            force,
        )
    )



import torch

from .abstract_model import AbstractModel
from .utils import ExtractLayerEmbeddings


class FFNN(AbstractModel):
    def __init__(
        self,
        output_dim: int,
        input_size: int,
        hidden_size: int,
        num_layers: int = 2,
        sigmoid: bool = False,
        softmax: bool = False,
        dropout: float = 0.5,
    ) -> None:
        """Feedforward neural network.

        Args:
            output_dim: Output dimension.
            input_size: Input size.
            hidden_size: Hidden size.
            num_layers: Number of layers.
            sigmoid: Whether to use sigmoid activation.
            softmax: Whether to use softmax activation.
            dropout: Dropout rate.
        """
        super().__init__(output_dim)
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.sigmoid = sigmoid
        self.softmax = softmax
        self.dropout = dropout

        layers = []
        layer_input = input_size
        for i in range(num_layers - 1):
            layers += [
                (f"Linear{i}", torch.nn.Linear(layer_input, hidden_size)),
                (f"ReLU{i}", torch.nn.ReLU()),
                (f"Dropout{i}", torch.nn.Dropout(dropout)),
            ]
            layer_input = hidden_size
        layers.append(
            (
                f"Linear{num_layers - 1}",
                torch.nn.Linear(layer_input, output_dim),
            )
        )
        if self.sigmoid:
            layers.append(("Sigmoid", torch.nn.Sigmoid()))
        if self.softmax:
            layers.append(("Softmax", torch.nn.Softmax(dim=1)))

        for name, layer in layers:
            self.add_module(name, layer)

        self._embedding_extractor = ExtractLayerEmbeddings(self)

    def embeddings(self, x: torch.Tensor) -> torch.Tensor:
        return self._embedding_extractor(x)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for layer in self.children():
            x = layer(x)
        return x



from abc import ABC, abstractmethod

import audobject
import torch


class AbstractModel(torch.nn.Module, audobject.Object, ABC):
    def __init__(self, output_dim: int) -> None:
        """Abstract model class.

        Args:
            output_dim: Output dimension of the model.
        """
        super().__init__()
        self.output_dim = output_dim

    @abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the model.

        Args:
            x: Input tensor.

        Returns:
            Output tensor.
        """

    @abstractmethod
    def embeddings(self, x: torch.Tensor) -> torch.Tensor:
        """Get embeddings from the model.

        Args:
            x: Input tensor.

        Returns:
            Embeddings.
        """



from .abstract_augmentation import AbstractAugmentation
from .augmentation_manager import AugmentationManager
from .augmentation_pipeline import AugmentationPipeline
from .augmentation_wrappers import (
    AlbumentationsAugmentation,
    AudiomentationsAugmentation,
    AugmentationWrapper,
    TorchaudioAugmentation,
    TorchAudiomentationsAugmentation,
    TorchvisionAugmentation,
)
from .choice_augmentation import Choice
from .image_augmentations import CutMix, MixUp, SampleGaussianWhiteNoise
from .sequential_augmentation import Sequential
from .spectrogram_augmentations import (
    FrequencyMask,
    GaussianNoise,
    SpecAugment,
    TimeMask,
    TimeShift,
    TimeWarp,
)


__all__ = [
    "AbstractAugmentation",
    "AlbumentationsAugmentation",
    "AudiomentationsAugmentation",
    "AugmentationManager",
    "AugmentationPipeline",
    "AugmentationWrapper",
    "Choice",
    "CutMix",
    "FrequencyMask",
    "GaussianNoise",
    "MixUp",
    "SampleGaussianWhiteNoise",
    "Sequential",
    "SpecAugment",
    "TimeMask",
    "TimeShift",
    "TimeWarp",
    "TorchaudioAugmentation",
    "TorchAudiomentationsAugmentation",
    "TorchvisionAugmentation",
]



import os
from typing import TYPE_CHECKING, Callable, List, Optional, Tuple, Type

import pandas as pd
import torch
from torch.utils.data import default_collate
from torchvision.transforms import v2

from .abstract_augmentation import AbstractAugmentation


if TYPE_CHECKING:
    from autrainer.datasets import AbstractDataset  # pragma: no cover


class BaseMixUpCutMix(AbstractAugmentation):
    def __init__(
        self,
        augmentation_class: Type,
        alpha: float = 1.0,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
    ) -> None:
        """Base class for MixUp and CutMix augmentations.

        Args:
            augmentation_class: The class of the augmentation to apply.
            alpha: Hyperparameter of the Beta distribution. Defaults to 1.0.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
        """
        super().__init__(order, p, generator_seed)
        self.augmentation_class = augmentation_class
        self.alpha = alpha

    def get_collate_fn(self, data: "AbstractDataset") -> Callable:
        if data.output_dim <= 1:
            raise ValueError(
                f"{self.augmentation_class.__name__} "
                "requires more than 1 class."
            )
        self.augmentation = self.augmentation_class(
            num_classes=data.output_dim, alpha=self.alpha
        )

        def _collate_fn(
            batch: List[Tuple[torch.Tensor, int, int]],
        ) -> List[Tuple[torch.Tensor, torch.Tensor, torch.Tensor]]:
            probability = torch.rand(1, generator=self.g).item()
            if probability < self.p:
                return self.augmentation(*default_collate(batch))

            batched = default_collate(batch)
            batched[1] = torch.nn.functional.one_hot(
                batched[1], data.output_dim
            ).float()
            return batched

        return _collate_fn

    def apply(self, x: torch.Tensor, index: int = None) -> torch.Tensor:
        return x


class MixUp(BaseMixUpCutMix):
    def __init__(
        self,
        alpha: float = 1.0,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
    ) -> None:
        """MixUp augmentation. As MixUp utilizes a collate function, the
        probability of applying the augmentation is drawn for each batch.

        Args:
            alpha: Hyperparameter of the Beta distribution. Defaults to 1.0.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
        """
        super().__init__(v2.MixUp, alpha, order, p, generator_seed)


class CutMix(BaseMixUpCutMix):
    def __init__(
        self,
        alpha: float = 1.0,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
    ) -> None:
        """CutMix augmentation. As CutMix utilizes a collate function, the
        probability of applying the augmentation is drawn for each batch.

        Args:
            alpha: Hyperparameter of the Beta distribution. Defaults to 1.0.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
        """
        super().__init__(v2.CutMix, alpha, order, p, generator_seed)


class SampleGaussianWhiteNoise(AbstractAugmentation):
    def __init__(
        self,
        snr_df: str,
        snr_col: str,
        sample_seed: int = None,
        order: int = 101,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
    ) -> None:
        """Sample-level gaussian white noise augmentation based on SNR values.

        Args:
            snr_df: Path to a CSV file containing SNR values for each sample.
                Index of the CSV file must match the index of the dataset.
            snr_col: Name of the column containing the SNR values.
            sample_seed: Seed for the random number generator used for sampling
                the noise. If a seed is provided, a consistent augmentation
                is applied to the same sample. Defaults to None.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 101.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
        """
        super().__init__(order, p, generator_seed)
        if not os.path.exists(snr_df):
            raise FileNotFoundError(f"File {snr_df} does not exist.")
        self.snr_df = snr_df
        self.snr_col = snr_col
        self.sample_seed = sample_seed
        df = pd.read_csv(self.snr_df)
        if self.snr_col not in df.columns:
            raise ValueError(
                f"Column {self.snr_col} not found in {self.snr_df}."
            )
        self.snr = df[self.snr_col]
        self._generator = torch.Generator()

    def apply(self, x: torch.Tensor, index: int = None) -> torch.Tensor:
        if self.sample_seed:
            self._generator.manual_seed(
                hash((self.sample_seed, index)) & 0xFFFFFFFF
            )

        snr = 10 ** (self.snr[index] / 10)
        energy = torch.mean(x**2)
        noise = torch.normal(0, 1, generator=self._generator, size=x.shape)
        noise_energy = torch.mean(noise**2)
        scale = torch.sqrt(energy / (snr * noise_energy))
        return x + noise * scale



from dataclasses import dataclass
import glob
import os
from typing import Optional

import autrainer

from .abstract_script import AbstractScript, MockParser
from .command_line_error import CommandLineError
from .utils import catch_cli_errors


@dataclass
class InferenceArgs:
    model: str
    input: str
    output: str
    checkpoint: str
    device: str
    extension: str
    recursive: bool
    embeddings: bool
    preprocess_cfg: Optional[str]
    window_length: Optional[float]
    stride_length: Optional[float]
    min_length: Optional[float]
    sample_rate: Optional[int]


class InferenceScript(AbstractScript):
    def __init__(self) -> None:
        e = (
            "Example: autrainer inference hf:autrainer/example "
            "input/ output/ -d cuda:0"
        )
        super().__init__(
            "inference",
            "Perform inference on a trained model.",
            epilog=e,
            dataclass=InferenceArgs,
        )

    def add_arguments(self) -> None:
        self.parser.add_argument(
            "model",
            type=str,
            help=(
                "Local path to model directory or Hugging Face link of the "
                "format: `hf:repo_id[@revision][:subdir]#local_dir`. "
                "Should contain at least one state subdirectory, the "
                "`model.yaml`, `file_handler.yaml`, `target_transform.yaml`, "
                "and `inference_transform.yaml` files."
            ),
        )
        self.parser.add_argument(
            "input",
            type=str,
            help=(
                "Path to input directory. Should contain audio files "
                "of the specified extension."
            ),
        )
        self.parser.add_argument(
            "output",
            type=str,
            help=(
                "Path to output directory. Output includes a YAML file with "
                "predictions and a CSV file with model outputs."
            ),
        )
        self.parser.add_argument(
            "-c",
            "--checkpoint",
            type=str,
            metavar="C",
            required=False,
            default="_best",
            help=(
                "Checkpoint to use for evaluation. "
                "Defaults to '_best' (on validation set)."
            ),
        )
        self.parser.add_argument(
            "-d",
            "--device",
            type=str,
            metavar="D",
            required=False,
            default="cpu",
            help=(
                "CUDA-enabled device to use for processing. "
                "Defaults to 'cpu'."
            ),
        )
        self.parser.add_argument(
            "-e",
            "--extension",
            type=str,
            metavar="E",
            required=False,
            default="wav",
            help=(
                "Type of file to look for in the input directory. "
                "Defaults to 'wav'."
            ),
        )
        self.parser.add_argument(
            "-r",
            "--recursive",
            action="store_true",
            help=(
                "Recursively search for files in the input directory. "
                "Defaults to False."
            ),
        )
        self.parser.add_argument(
            "-emb",
            "--embeddings",
            action="store_true",
            help=(
                "Extract embeddings from the model in addition to predictions."
                "For each file, a .pt file with embeddings will be saved."
                "Defaults to False."
            ),
        )

        self.parser.add_argument(
            "-p",
            "--preprocess-cfg",
            type=str,
            metavar="P",
            required=False,
            default="default",
            help=(
                "Preprocessing configuration to apply to input. Can be a path "
                "to a YAML file or the name of the preprocessing configuration "
                "in the local or autrainer 'conf/preprocessing' directory. "
                "If 'default', the default preprocessing configuration used "
                "during training will be applied. If 'None', no preprocessing "
                "will be applied. Defaults to 'default'."
            ),
        )
        self.parser.add_argument(
            "-w",
            "--window-length",
            type=float,
            metavar="W",
            required=False,
            default=None,
            help=(
                "Window length for sliding window inference in seconds. "
                "If None, the entire input will be processed at once. "
                "Defaults to None."
            ),
        )
        self.parser.add_argument(
            "-s",
            "--stride-length",
            type=float,
            metavar="S",
            required=False,
            default=None,
            help=(
                "Stride length for sliding window inference in seconds. "
                "If None, the entire input will be processed at once. "
                "Defaults to None."
            ),
        )
        self.parser.add_argument(
            "-m",
            "--min-length",
            type=float,
            metavar="M",
            required=False,
            default=None,
            help=(
                "Minimum length of audio file to process in seconds. "
                "Files shorter than the minimum length are padded with zeros. "
                "If None, no minimum length is enforced. Defaults to None."
            ),
        )
        self.parser.add_argument(
            "-sr",
            "--sample-rate",
            type=int,
            metavar="SR",
            required=False,
            default=None,
            help=(
                "Sample rate of audio files in Hz. Has to be specified for "
                "sliding window inference. Defaults to None."
            ),
        )

    def main(self, args: InferenceArgs) -> None:
        args.model = self._assert_model_exists(args)
        self._assert_valid_input(args)
        self._assert_valid_device(args)
        self._assert_valid_checkpoint(args)
        self._assert_and_set_preprocess_path(args)
        self._inference(args)

    def _assert_model_exists(self, args: InferenceArgs) -> str:
        from autrainer.serving import get_model_path

        try:
            return get_model_path(args.model)
        except ValueError as e:
            raise CommandLineError(self.parser, str(e), code=1)

    def _assert_checkpoint_exists(self, args: InferenceArgs) -> None:
        path = os.path.join(args.model, args.checkpoint, "model.pt")
        if os.path.exists(path):
            return
        raise CommandLineError(
            self.parser,
            f"Checkpoint '{args.checkpoint}' does not exist.",
            code=1,
        )

    def _assert_valid_input(self, args: InferenceArgs) -> None:
        if not os.path.exists(args.input):
            raise CommandLineError(
                self.parser,
                f"Input '{args.input}' does not exist.",
                code=1,
            )
        if not os.path.isdir(args.input):
            raise CommandLineError(
                self.parser,
                f"Input '{args.input}' is not a directory.",
                code=1,
            )

        pattern = (
            f"**/*.{args.extension}"
            if args.recursive
            else f"*.{args.extension}"
        )
        if not glob.glob(
            os.path.join(args.input, pattern),
            recursive=True,
        ):
            raise CommandLineError(
                self.parser,
                f"No '{args.extension}' files found in '{args.input}'.",
                code=1,
            )

    def _assert_valid_device(self, args: InferenceArgs) -> None:
        if args.device == "cpu" or args.device.startswith("cuda:"):
            return
        raise CommandLineError(
            self.parser,
            (
                f"Invalid device '{args.device}'. "
                "Please specify 'cpu' or 'cuda:<device_id>'."
            ),
        )

    def _assert_valid_checkpoint(self, args: InferenceArgs) -> None:
        path = os.path.join(args.model, args.checkpoint, "model.pt")
        if os.path.exists(path):
            return
        raise CommandLineError(
            self.parser,
            f"Checkpoint '{args.checkpoint}' does not exist.",
            code=1,
        )

    def _assert_and_set_preprocess_path(self, args: InferenceArgs) -> None:
        if args.preprocess_cfg == "default":
            return

        if args.preprocess_cfg == "None" or args.preprocess_cfg is None:
            args.preprocess_cfg = None
            return

        if not args.preprocess_cfg.endswith(".yaml"):
            args.preprocess_cfg += ".yaml"

        if os.path.exists(args.preprocess_cfg):
            return

        local_path = os.path.join("conf", "preprocessing", args.preprocess_cfg)
        if os.path.exists(local_path):
            args.preprocess_cfg = local_path
            return

        lib_path = os.path.join(
            os.path.dirname(autrainer.__path__[0]),
            "autrainer-configurations",
            "preprocessing",
            args.preprocess_cfg,
        )
        if os.path.exists(lib_path):
            args.preprocess_cfg = lib_path
            return

        raise CommandLineError(
            self.parser,
            (
                f"Preprocessing configuration '{args.preprocess_cfg}' "
                "does not exist."
            ),
            code=1,
        )

    def _inference(self, args: InferenceArgs) -> None:
        from autrainer.serving import Inference

        inference = Inference(
            model_path=args.model,
            checkpoint=args.checkpoint,
            device=args.device,
            preprocess_cfg=args.preprocess_cfg,
            window_length=args.window_length,
            stride_length=args.stride_length,
            min_length=args.min_length,
            sample_rate=args.sample_rate,
        )

        results = inference.predict_directory(
            args.input,
            extension=args.extension,
            recursive=args.recursive,
        )
        inference.save_prediction_yaml(results, args.output)
        inference.save_prediction_results(results, args.output)

        if not args.embeddings:
            return

        results = inference.embed_directory(
            args.input,
            extension=args.extension,
            recursive=args.recursive,
        )
        inference.save_embeddings(
            results,
            args.output,
            input_extension=args.extension,
        )


@catch_cli_errors
def inference(
    model: str,
    input: str,
    output: str,
    checkpoint: str = "_best",
    device: str = "cpu",
    extension: str = "wav",
    recursive: bool = False,
    embeddings: bool = False,
    preprocess_cfg: Optional[str] = "default",
    window_length: Optional[float] = None,
    stride_length: Optional[float] = None,
    min_length: Optional[float] = None,
    sample_rate: Optional[int] = None,
) -> None:
    """Perform inference on a trained model.

    If called in a notebook, the function will not raise an error and print
    the error message instead.

    Args:
        model: Local path to model directory or Hugging Face link of the
            format: `hf:repo_id[@revision][:subdir]#local_dir`.
            Should contain at least one state subdirectory, the
            `model.yaml`, `file_handler.yaml`, `target_transform.yaml`,
            and `inference_transform.yaml` files.
        input: Path to input directory. Should contain audio files of the
            specified extension.
        output: Path to output directory. Output includes a YAML file with
            predictions and a CSV file with model outputs.
        checkpoint: Checkpoint to use for evaluation. Defaults to '_best'.
        device: CUDA-enabled device to use for processing. Defaults to 'cpu'.
        extension: Type of file to look for in the input directory.
            Defaults to 'wav'.
        recursive: Recursively search for files in the input directory.
            Defaults to False.
        embeddings: Extract embeddings from the model in addition to
            predictions. For each file, a .pt file with embeddings will be
            saved. Defaults to False.
        preprocess_cfg: Preprocessing configuration to apply to input. Can be
            a path to a YAML file or the name of the preprocessing
            configuration in the local or autrainer 'conf/preprocessing'
            directory. If "default", the default preprocessing configuration
            used during training will be applied. If None, no preprocessing
            will be applied. Defaults to "default".
        window_length: Window length for sliding window inference in seconds.
            If None, the entire input will be processed at once.
            Defaults to None.
        stride_length: Stride length for sliding window inference in seconds.
            If None, the entire input will be processed at once.
            Defaults to None.
        min_length: Minimum length of audio file to process in seconds.
            Files shorter than the minimum length are padded with zeros.
            If None, no minimum length is enforced. Defaults to None.
        sample_rate: Sample rate of audio files in Hz. Has to be specified for
            sliding window inference. Defaults to None.

    Raises:
        CommandLineError: If the model, input, or preprocessing configuration
            does not exist, or if the device is invalid.
    """
    script = InferenceScript()
    script.parser = MockParser()
    script.main(
        InferenceArgs(
            model,
            input,
            output,
            checkpoint,
            device,
            extension,
            recursive,
            embeddings,
            preprocess_cfg,
            window_length,
            stride_length,
            min_length,
            sample_rate,
        )
    )



from abc import ABC, abstractmethod
from typing import Any, List, Union

import audobject
import numpy as np
import torch


class AbstractTargetTransform(ABC, audobject.Object):
    def __init__(self) -> None:
        """Abstract target transform for handling target or label
        transformations in a dataset.

        Serves as the base for creating custom target transforms that handle
        encoding and decoding targets or labels of a dataset, obtaining
        predictions from a batch of model outputs, and determining the majority
        vote from a list of targets.
        """

    def __call__(self, x: Any) -> Any:
        return self.encode(x)

    @abstractmethod
    def encode(self, x: Any) -> Any:
        """Encode a target or label.

        Args:
            x: Target or label.

        Returns:
            Encoded target or label.
        """

    @abstractmethod
    def decode(self, x: Any) -> Any:
        """Decode a target or label. Serve as the inverse operation of encode.

        Args:
            x: Encoded target or label.

        Returns:
            Decoded target or label.
        """

    @abstractmethod
    def predict_batch(self, x: torch.Tensor) -> Union[List[Any], Any]:
        """Get encoded predictions from a batch of model outputs.

        Args:
            x: Batch of model outputs.

        Returns:
            Encoded predictions.
        """

    @abstractmethod
    def majority_vote(self, x: List[Any]) -> Any:
        """Get the majority vote from a list of decoded targets or labels.

        The majority vote is defined by the subclasses and may be the most
        frequent target, the average of all targets, or any other operation
        that determines a majority vote based on the list of targets or labels.

        Args:
            x: List of decoded targets or labels.

        Returns:
            Decoded majority vote.
        """


class MinMaxScaler(AbstractTargetTransform):
    def __init__(self, minimum: float, maximum: float) -> None:
        """Minimum-Maximum Scaler for regression targets.

        Args:
            minimum: Minimum value of all target values.
            maximum: Maximum value of all target values.

        Raises:
            ValueError: If minimum is not less than maximum.
        """
        if not minimum < maximum:
            raise ValueError(
                f"Minimum '{minimum}' must be less than maximum '{maximum}'."
            )
        self.minimum = float(minimum)
        self.maximum = float(maximum)

    def encode(self, x: float) -> float:
        """Encode a target value by scaling it between the minimum and maximum.

        Args:
            x: Target value.

        Returns:
            Scaled target value.
        """
        return (x - self.minimum) / (self.maximum - self.minimum)

    def decode(self, x: float) -> float:
        """Decode a target value by reversing the scaling between the minimum
        and maximum. Inverse operation of encode.

        Args:
            x: Scaled target value.

        Returns:
            Unscaled target value.
        """
        return x * (self.maximum - self.minimum) + self.minimum

    def predict_batch(self, x: torch.Tensor) -> Union[List[float], float]:
        """Get encoded predictions from a batch of model outputs by
        squeezing the tensor and converting it to a list.


        Args:
            x: Batch of model outputs.

        Returns:
            Encoded predictions.
        """
        return x.squeeze().tolist()

    def majority_vote(self, x: List[float]) -> float:
        """Get the majority vote from a list of target values by averaging
        the predictions.

        Args:
            x: List of target values.

        Returns:
            Average target value.
        """
        return sum(x) / len(x)


class MultiLabelEncoder(AbstractTargetTransform):
    def __init__(self, threshold: float, labels: List[str]) -> None:
        """Multi-label encoder for multi-label classification targets.

        Args:
            threshold: Class-wise prediction threshold.
            labels: List of target labels.

        Raises:
            ValueError: If threshold is not between 0 and 1.
        """
        if not 0 <= threshold <= 1:
            raise ValueError(
                f"Threshold '{threshold}' must be between 0 and 1."
            )
        self.threshold = threshold
        self.labels = labels

    def encode(self, x: Union[List[int], List[str]]) -> torch.Tensor:
        """Encode a list of target labels by creating a binary tensor where
        each element is 1 if the label is in the list of target labels and 0
        otherwise.

        If the input is already a list of integers, it is not encoded.

        Args:
            x: List of target labels or list of integers.

        Returns:
            Binary tensor of encoded target labels.
        """
        if isinstance(x, (torch.Tensor, np.ndarray)) or all(
            isinstance(i, (int, float)) for i in x
        ):
            return torch.Tensor(x)
        return torch.Tensor([label in x for label in self.labels])

    def decode(self, x: List[int]) -> List[str]:
        """Decode a binary tensor of encoded target labels to a list of target
        labels.

        Args:
            x: Binary tensor of encoded target labels.

        Returns:
            List of target labels.
        """
        return [label for i, label in enumerate(self.labels) if x[i]]

    def predict_batch(
        self, x: torch.Tensor
    ) -> Union[List[List[int]], List[int]]:
        """Predict the binary tensor of encoded predictions by thresholding the
        model outputs.

        Args:
            x: Batch of model outputs.

        Returns:
            Binary tensor of encoded predictions.
        """
        return (torch.sigmoid(x) > self.threshold).int().squeeze().tolist()

    def majority_vote(self, x: List[List[str]]) -> List[str]:
        """Get the majority vote from a list of lists of decoded target labels
        for each label. If a label is predicted by at least half of the
        predictions, it is included in the majority vote.

        Args:
            x: List of lists of decoded target labels.

        Returns:
            List of target labels in the majority vote.
        """
        x = np.array([self.encode(i).tolist() for i in x])
        x = np.mean(x, axis=0).round().astype(int).tolist()
        return self.decode(x)


class LabelEncoder(AbstractTargetTransform):
    def __init__(self, labels: List[str]) -> None:
        """Label encoder for single-label classification targets.

        Args:
            labels: List of target labels.
        """
        self.labels = sorted(labels)
        codes = range(len(self.labels))
        self.inverse_map = {
            code: label for code, label in zip(codes, self.labels)
        }
        self.map = {label: code for code, label in zip(codes, self.labels)}

    def encode(self, x: str) -> int:
        """Encode a target label by mapping it to an integer.

        Args:
            x: Target label.

        Returns:
            Encoded target label.
        """
        return self.map[x]

    def decode(self, x: int) -> str:
        """Decode an encoded target label integer by mapping it to a label.

        Args:
            x: Encoded target label.

        Returns:
            Decoded target label.
        """
        return self.inverse_map[x]

    def predict_batch(self, x: torch.Tensor) -> Union[List[int], int]:
        """Get the encoded predictions from a batch of model outputs by
        obtaining the index of the maximum value.

        Args:
            x: Batch of model outputs.

        Returns:
            Encoded predictions.
        """
        return x.argmax(-1).squeeze().tolist()

    def majority_vote(self, x: List[str]) -> str:
        """Get the majority vote from a list of decoded labels by
        determining the most frequently predicted label.

        Args:
            x: List of decoded labels.

        Returns:
            Decoded majority vote.
        """
        return max(set(x), key=x.count)



from pathlib import Path
from typing import List, Optional, Tuple, Union

import pandas as pd
import torch

from autrainer.transforms import SmartCompose

from .file_handlers import AbstractFileHandler
from .target_transforms import AbstractTargetTransform


class DatasetWrapper(torch.utils.data.Dataset):
    def __init__(
        self,
        path: str,
        features_subdir: str,
        index_column: str,
        target_column: Union[str, List[str]],
        file_type: str,
        file_handler: AbstractFileHandler,
        df: pd.DataFrame,
        transform: Optional[SmartCompose] = None,
        target_transform: Optional[AbstractTargetTransform] = None,
    ):
        """Wrapper around torch.utils.data.Dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            df: Dataframe containing the index and target column(s).
            transform: Transform to apply to the features. Defaults to None.
            target_transform: Target transform to apply to the target.
                Defaults to None.

        """
        self.path = path
        self.features_subdir = features_subdir
        self.index_column = index_column
        self.target_column = target_column
        self.file_type = file_type
        self.file_handler = file_handler
        self.df = df.copy()
        self.transform = transform
        self.target_transform = target_transform

        self.df[self.index_column] = self.df[self.index_column].apply(
            self._create_file_path
        )

    def _create_file_path(self, file: str) -> str:
        path = Path(self.path, self.features_subdir, file)
        path = path.with_suffix(f".{self.file_type}")
        return str(path)

    def __len__(self) -> int:
        return len(self.df)

    def __getitem__(
        self,
        item: int,
    ) -> Tuple[torch.Tensor, Union[int, torch.Tensor], int]:
        """Get item from the dataset.

        Args:
            item: Index of the item.

        Returns:
            Tuple containing the data, target and item index.
        """
        index = self.df.index[item]
        item_path = self.df.loc[index, self.index_column]
        data = self.file_handler.load(item_path)
        target = self.df.loc[index, self.target_column]
        if isinstance(target, pd.Series):
            target = torch.Tensor(target.to_list())

        if self.transform is not None:
            data = self.transform(data, index=item).float()

        if self.target_transform is not None:
            target = self.target_transform(target)

        return data, target, item



from typing import TYPE_CHECKING

import torch


if TYPE_CHECKING:
    from autrainer.datasets import AbstractDataset  # pragma: no cover


class CrossEntropyLoss(torch.nn.CrossEntropyLoss):
    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Wrapper for `torch.nn.CrossEntropyLoss.forward`.

        Converts the targets to `long` if it is a 1D tensor.

        Args:
            x: Batched model outputs.
            y: Targets.

        Returns:
            Loss.
        """
        if y.ndim == 1:
            y = y.long()
        return super().forward(x, y)


class BalancedCrossEntropyLoss(CrossEntropyLoss):
    def setup(self, data: "AbstractDataset") -> None:
        """Calculate balanced weights for the dataset based on the target
        frequency in the training set.

        Args:
            data: Instance of the dataset.
        """
        frequency = (
            data.df_train[data.target_column]
            .map(data.target_transform)
            .value_counts()
            .sort_index()
            .values
        )
        weight = torch.tensor(1 / frequency, dtype=torch.float32)
        weight /= weight.sum()
        self.weight = weight


class MSELoss(torch.nn.MSELoss):
    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Wrapper for `torch.nn.MSELoss.forward`.

        Squeezes the model outputs along the last dimension to match the shape
        of the targets.
        Converts the targets to `float`.

        Args:
            x: Batched model outputs.
            y: Targets.

        Returns:
            Loss.
        """
        # TODO: is there a more elegant way to handle broadcasting
        return super().forward(x.squeeze(-1), y.float())



from collections import defaultdict
import os
import shutil
from typing import List, Optional

from omegaconf import DictConfig, OmegaConf
import pandas as pd

import autrainer
from autrainer.core.constants import NAMING_CONVENTION, VALID_AGGREGATIONS
from autrainer.core.plotting import PlotMetrics
from autrainer.core.utils import Timer
from autrainer.loggers import AbstractLogger
from autrainer.metrics import AbstractMetric

from .postprocessing_utils import (
    get_plotting_params,
    get_run_names,
    get_training_type,
    load_yaml,
    save_yaml,
)
from .summarize import SummarizeGrid


class AggregateGrid:
    def __init__(
        self,
        results_dir: str,
        experiment_id: str,
        aggregate_list: List[str],
        aggregate_prefix: str = "agg",
        training_dir: str = "training",
        max_runs_plot: Optional[int] = None,
        aggregate_name: Optional[str] = None,
        aggregated_dict: Optional[dict] = None,
        plot_params: Optional[dict] = None,
    ) -> None:
        """Aggregate the results of a grid search over one or more parameters.

        If loggers have been used for the grid search, the aggregated results
        will be logged to the same loggers.

        Args:
            results_dir: The directory where the results are stored.
            experiment_id: The ID of the grid search experiment.
            aggregate_list: The list of parameters to aggregate over.
            aggregate_prefix: The prefix for the aggregated experiment ID.
                Defaults to "agg".
            training_dir: The directory of the training results of the
                experiment. Defaults to "training".
            max_runs_plot: The maximum number of best runs to plot. If None,
                all runs will be plotted. Defaults to None.
            aggregate_name: The name of the aggregated experiment. If None,
                it will be generated from the aggregate_list. Defaults to None.
            aggregated_dict: A dictionary mapping the aggregated experiment
                names to the runs to aggregate. If None, the runs will be
                aggregated based on the aggregate_list. Defaults to None.
            plot_params: Additional parameters for plotting. Defaults to None.
        """
        self.aggregate_list = aggregate_list
        self.results_dir = results_dir
        self.experiment_id = experiment_id
        self.aggregate_name = aggregate_name or "_".join(
            (aggregate_prefix, *self.aggregate_list)
        )
        self.aggregated_dict = aggregated_dict
        self.plot_params = plot_params
        self.output_directory = os.path.join(
            self.results_dir, self.experiment_id, self.aggregate_name
        )
        self.training_directory = os.path.join(
            self.results_dir, self.experiment_id, training_dir
        )
        self.max_runs_plot = max_runs_plot
        if os.path.exists(self.output_directory):
            shutil.rmtree(self.output_directory)
        os.makedirs(self.output_directory, exist_ok=True)
        self.run_names = get_run_names(self.training_directory)
        self.training_type = get_training_type(
            self.training_directory, self.run_names
        )
        self.run_names.sort()
        if self.plot_params is None:
            self.plot_params = get_plotting_params(
                self.training_directory, self.run_names[0]
            )

    def aggregate(self) -> None:
        """Aggregate the runs based on the specified parameters."""
        if self.aggregated_dict is not None:
            aggregated_runs = self.aggregated_dict
        else:
            aggregated_runs = self._aggregate_run_names(self.aggregate_list)
        for agg_name, run_list in aggregated_runs.items():
            self._aggregate_best(agg_name, run_list)
            self._aggregate_test(agg_name, run_list)
            self._aggregate_config(agg_name, run_list)
            self._aggregate_timer(agg_name, run_list)
            self._aggregate_metrics(agg_name, run_list)
            save_yaml(
                os.path.join(self.output_directory, agg_name, "runs.yaml"),
                run_list,
            )

    def summarize(self) -> None:
        """Summarize the aggregated results."""
        sg = SummarizeGrid(
            results_dir=self.results_dir,
            experiment_id=self.experiment_id,
            training_dir=self.aggregate_name,
            summary_dir=self.aggregate_name,
            clear_old_outputs=False,
            training_type=self.training_type,
            max_runs_plot=self.max_runs_plot,
            plot_params=self.plot_params,
        )
        sg.summarize()
        sg.plot_metrics()

    def _check_if_valid_aggregation(self, over: list) -> None:
        for o in over:
            if o not in VALID_AGGREGATIONS:
                raise ValueError(f"Invalid aggregation: {o}")

    def _aggregate_run_names(self, over: list) -> dict:
        self._check_if_valid_aggregation(over)
        param_dict = {p: i for i, p in enumerate(NAMING_CONVENTION)}
        over_idxs = [param_dict[p] for p in over]
        aggregated = defaultdict(list)
        for run_name in self.run_names:
            params = run_name.split("_")
            for idx in over_idxs:
                params[idx] = "#"
            agg_key = "_".join(params)
            aggregated[agg_key].append(run_name)
        return aggregated

    def _aggregate_best(self, agg_name: str, run_list: list):
        os.makedirs(
            os.path.join(self.output_directory, agg_name, "_best"),
            exist_ok=True,
        )
        metrics = self._aggregate_yaml(run_list, "_best/dev.yaml", "dev")
        save_yaml(
            os.path.join(self.output_directory, agg_name, "_best", "dev.yaml"),
            metrics,
        )

    def _aggregate_test(self, agg_name: str, run_list: list):
        path = os.path.join(self.output_directory, agg_name, "_test")
        os.makedirs(path, exist_ok=True)
        metrics = self._aggregate_yaml(
            run_list, "_test/test_holistic.yaml", "test"
        )
        save_yaml(os.path.join(path, "test_holistic.yaml"), metrics)

    def _aggregate_yaml(self, run_list: list, path: str, yaml_type: str):
        assert yaml_type in ["dev", "test"]
        loss_type = "dev_loss" if yaml_type == "dev" else "loss"
        dfs = []
        for run in run_list:
            metrics = load_yaml(
                os.path.join(self.training_directory, run, path)
            )
            dfs.append(pd.DataFrame(metrics))
        df = pd.concat(dfs, keys=run_list, names=["run", "type"])
        df_mean = df.groupby(level="type").mean().reset_index()
        df_std = df.groupby(level="type").std().fillna(0).reset_index()
        df_std["type"] = df_std["type"].apply(lambda x: f"{x}.std")
        df = pd.concat([df_mean, df_std]).set_index(["type"])
        metrics = df.to_dict()
        metrics[loss_type] = {
            k: v for k, v in metrics[loss_type].items() if "all" == k
        }
        if yaml_type == "dev":
            metrics["iteration"] = {
                k: v for k, v in metrics["iteration"].items() if "all" == k
            }
        return metrics

    def _aggregate_config(self, agg_name: str, run_list: list):
        path = os.path.join(self.output_directory, agg_name, ".hydra")
        os.makedirs(path, exist_ok=True)
        runs = [
            OmegaConf.load(
                os.path.join(
                    self.training_directory, r, ".hydra", "config.yaml"
                )
            )
            for r in run_list
        ]
        config = runs.pop(0)
        if self.aggregated_dict is None:
            for key in self.aggregate_list:
                config[key] = self._replace_differing_values(config, runs, key)
        save_yaml(
            os.path.join(path, "config.yaml"), OmegaConf.to_container(config)
        )

    def _replace_differing_values(
        self,
        config: DictConfig,
        runs: List[DictConfig],
        key: str,
    ) -> DictConfig:
        if not runs:
            return config[key]

        if not isinstance(config[key], DictConfig):
            return "#"

        for k in config[key].keys():
            if not isinstance(config[key][k], DictConfig):
                values = [r[key][k] for r in runs + [config]]
                if len(set(values)) > 1:
                    config[key][k] = "#"
            else:
                config[key][k] = self._replace_differing_values(
                    config[key], [r[key] for r in runs], k
                )

        return config[key]

    def _aggregate_timer(self, agg_name: str, run_list: list):
        mean_timer = {
            "train": {"mean_seconds": 0, "total_seconds": 0},
            "dev": {"mean_seconds": 0, "total_seconds": 0},
            "test": {"mean_seconds": 0, "total_seconds": 0},
        }
        for run in run_list:
            timers = load_yaml(
                os.path.join(self.training_directory, run, "timer.yaml")
            )
            for k, v in timers.items():
                mean_timer[k]["mean_seconds"] += v["mean_seconds"]
                mean_timer[k]["total_seconds"] += v["total_seconds"]
        for k, v in mean_timer.items():
            v["mean_seconds"] /= len(run_list)
            v["mean"] = Timer.pretty_time(v["mean_seconds"])
            v["total_seconds"] /= len(run_list)
            v["total"] = Timer.pretty_time(v["total_seconds"])
        save_yaml(
            os.path.join(self.output_directory, agg_name, "timer.yaml"),
            mean_timer,
        )

    def _aggregate_metrics(self, agg_name: str, run_list: list):
        dfs = []
        for run in run_list:
            df = pd.read_csv(
                os.path.join(self.training_directory, run, "metrics.csv"),
                index_col="iteration",
            )
            dfs.append(df)
        cfg = OmegaConf.load(
            os.path.join(
                self.output_directory, agg_name, ".hydra", "config.yaml"
            )
        )
        if isinstance(cfg.dataset.metrics, str):
            raise ValueError(
                "Unable to aggregate over datasets with different metrics."
            )
        metric_fns = [
            autrainer.instantiate_shorthand(m, instance_of=AbstractMetric)
            for m in cfg.dataset.metrics
        ]
        tracking_metric = autrainer.instantiate_shorthand(
            config=cfg.dataset.tracking_metric,
            instance_of=AbstractMetric,
        )

        df = pd.concat(dfs, keys=run_list, names=["run", "iteration"])
        mean_df = df.groupby(level="iteration").mean()
        std_df = df.groupby(level="iteration").std().fillna(0)
        df = mean_df.join(std_df, rsuffix=".std")
        df["iteration"] = df.index
        df.to_csv(
            os.path.join(self.output_directory, agg_name, "metrics.csv"),
            index=False,
        )
        df.drop(columns=["iteration"], inplace=True)

        plotter = PlotMetrics(
            output_directory=os.path.join(self.output_directory, agg_name),
            training_type=self.training_type,
            metric_fns=metric_fns,
            **self.plot_params,
        )
        plotter.plot_run(df)

        loggers = []
        for logger in cfg.get("loggers", []):
            loggers.append(
                autrainer.instantiate_shorthand(
                    logger,
                    instance_of=AbstractLogger,
                    exp_name=self.experiment_id + "." + self.aggregate_name,
                    run_name=agg_name,
                    metrics=metric_fns,
                    tracking_metric=tracking_metric,
                )
            )
        timers = load_yaml(
            os.path.join(self.output_directory, agg_name, "timer.yaml")
        )
        test_metrics = load_yaml(
            os.path.join(
                self.output_directory, agg_name, "_test", "test_holistic.yaml"
            )
        )
        for logger in loggers:
            logger.setup()
            logger.log_params(cfg)
            logger.log_timers(
                {
                    "time.train.mean": timers["train"]["mean"],
                    "time.dev.mean": timers["dev"]["mean"],
                    "time.test.mean": timers["test"]["mean"],
                }
            )
            for iteration in df.index:
                metrics = df.loc[iteration].to_dict()
                metrics = {
                    k: v for k, v in metrics.items() if not k.endswith(".std")
                }
                logger.log_metrics(metrics, iteration)
            logger.log_metrics(
                {"test_" + k: v["all"] for k, v in test_metrics.items()}
            )
            logger.log_artifact(
                os.path.join(agg_name, ".hydra", "config.yaml"),
                self.output_directory,
            )
            logger.log_artifact(
                os.path.join(agg_name, "metrics.csv"), self.output_directory
            )
            logger.end_run()



from abc import ABC, abstractmethod
from typing import Optional

import audobject
import torch


class AbstractTransform(ABC, audobject.Object):
    def __init__(self, order: int = 0, **kwargs) -> None:
        """Abstract class for a transform.

        Args:
            order: The order of the transform in the pipeline. Larger means
                later in the pipeline. If multiple transforms have the same
                order, they are applied in the order they were added to the
                pipeline. Defaults to 0.
            kwargs: Additional keyword arguments to store in the object.
        """
        super().__init__(**kwargs)
        self.order = order
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self) -> str:
        args_dict = {}
        for k, v in self.__dict__.items():
            if k.startswith("_"):
                continue
            args_dict[k] = v
        args_repr = ", ".join(f"{k}={v!r}" for k, v in args_dict.items())
        return self.__class__.__name__ + f"({args_repr})"

    @abstractmethod
    def __call__(
        self,
        x: torch.Tensor,
        index: Optional[int] = None,
    ) -> torch.Tensor:
        """Apply the transform to the input tensor.

        Args:
            x: The input tensor.
            index: The index of the input tensor in the dataset.
                Defaults to None.

        Returns:
            The transformed tensor.
        """



import audmetric

from .abstract_metric import BaseAscendingMetric, BaseDescendingMetric


class MSE(BaseDescendingMetric):
    def __init__(self):
        """Mean squared error metric using `audmetric.mean_squared_error`."""
        super().__init__(name="mse", fn=audmetric.mean_squared_error)


class MAE(BaseDescendingMetric):
    def __init__(self):
        """Mean absolute error metric using `audmetric.mean_absolute_error`."""
        super().__init__(name="mae", fn=audmetric.mean_absolute_error)


class PCC(BaseAscendingMetric):
    def __init__(self):
        """Pearson correlation coefficient metric using
        `audmetric.pearson_cc`.
        """
        super().__init__(name="pcc", fn=audmetric.pearson_cc)


class CCC(BaseAscendingMetric):
    def __init__(self):
        """Concordance correlation coefficient metric using
        `audmetric.concordance_cc`.
        """
        super().__init__(name="ccc", fn=audmetric.concordance_cc)



import logging
import os
import sys
from typing import TYPE_CHECKING, Dict, List, Optional, Union
import warnings

import audobject
import numpy as np
import pandas as pd
import torch
from torchinfo import summary
import yaml

from autrainer.metrics import AbstractMetric


if TYPE_CHECKING:
    from autrainer.datasets import AbstractDataset  # pragma: no cover


class Bookkeeping:
    def __init__(
        self,
        output_directory: str,
        file_handler_path: Optional[str] = None,
    ) -> None:
        """Bookkeeping to handle general disk operations and interactions.

        Args:
            output_directory: Output directory to save files to.
            file_handler_path: Path to save the log file to. Defaults to None.
        """
        self.output_directory = output_directory
        os.makedirs(self.output_directory, exist_ok=True)
        self.original_stdout = sys.stdout
        # ? Setup Custom Logging
        self.logger = logging.getLogger()
        if not self.logger.hasHandlers() or file_handler_path is not None:
            self._setup_logger(file_handler_path)

        for handler in self.logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                handler.setFormatter(
                    logging.Formatter(
                        "[%(asctime)s][%(levelname)s]\n%(message)s\n"
                    )
                )

    def _setup_logger(self, fp: Optional[str] = None) -> None:
        self.logger.setLevel(logging.INFO)
        if not self.logger.hasHandlers():
            self.logger.addHandler(logging.StreamHandler())
        if fp is not None:
            self.logger.addHandler(logging.FileHandler(fp))
        else:
            self.logger.addHandler(
                logging.FileHandler(
                    os.path.join(self.output_directory, "bookkeeping.log")
                )
            )

    def log(self, message: str, level: int = logging.INFO) -> None:
        """Log a message.

        Args:
            message: Message to log.
            level: Logging level. Defaults to logging.INFO.
        """
        self.logger.log(level, message)

    def log_to_file(self, message: str, level: int = logging.INFO) -> None:
        """Log a message to the file handler.

        Args:
            message: Message to log.
            level: Logging level. Defaults to logging.INFO.
        """
        for handler in self.logger.handlers:
            if isinstance(handler, logging.FileHandler):
                handler.emit(
                    logging.LogRecord(
                        self.logger.name,
                        level,
                        None,
                        None,
                        message,
                        None,
                        None,
                    )
                )

    def create_folder(self, folder_name: str, path: str = "") -> None:
        """Create a new folder in the output directory.

        Args:
            folder_name: Name of the folder to create.
            path: Subdirectory to create the folder in. Defaults to "".
        """
        os.makedirs(
            os.path.join(self.output_directory, path, folder_name),
            exist_ok=True,
        )

    def save_model_summary(
        self,
        model: torch.nn.Module,
        dataset: "AbstractDataset",
        filename: str,
    ) -> None:
        """Save a model summary to a file.

        Args:
            model: Model to summarize.
            dataset: Dataset to get the input size from.
            filename: Name of the file to save the summary to.
        """
        x = np.expand_dims(dataset[0][0], axis=0).shape
        with open(
            os.path.join(self.output_directory, filename),
            "w",
            encoding="utf-8",
        ) as f:
            sys.stdout = f
            s = summary(
                model=model,
                input_size=(x),
                col_names=[
                    "input_size",
                    "output_size",
                    "num_params",
                    "trainable",
                ],
                col_width=20,
                row_settings=["var_names"],
            )
            sys.stdout = self.original_stdout
        model_summary = {
            "total_mult_adds": s.total_mult_adds,
            "total_output_bytes": s.total_output_bytes,
            "total_params": s.total_params,
            "trainable_params": s.trainable_params,
            "total_param_bytes": s.total_param_bytes,
        }
        with open(
            os.path.join(
                self.output_directory, filename.replace(".txt", ".yaml")
            ),
            "w",
        ) as f:
            yaml.dump(model_summary, f)

    def save_state(
        self,
        obj: Union[
            torch.nn.Module,
            torch.optim.Optimizer,
            torch.optim.lr_scheduler.LRScheduler,
        ],
        filename: str,
        path: str = "",
    ) -> None:
        """Save the state of an object.

        Args:
            obj: Object to save the state of.
            filename: Name of the file to save the state to.
            path: Subdirectory to save the state to. Defaults to "".

        Raises:
            TypeError: If the object type is not supported.
        """
        p = os.path.join(self.output_directory, path, filename)
        _i = (
            torch.nn.Module,
            torch.optim.Optimizer,
            torch.optim.lr_scheduler.LRScheduler,
        )
        if not isinstance(obj, _i):
            raise TypeError(
                f"save_state of type {type(obj)} is not supported."
            )

        os.makedirs(os.path.join(self.output_directory, path), exist_ok=True)
        torch.save(obj.state_dict(), p)

    def load_state(
        self,
        obj: Union[
            torch.nn.Module,
            torch.optim.Optimizer,
            torch.optim.lr_scheduler.LRScheduler,
        ],
        filename: str,
        path: str = "",
    ) -> None:
        """Load the state of an object.

        Args:
            obj: Object to load the state into.
            filename: Name of the file to load the state from.
            path: Subdirectory to load the state from. Defaults to "".

        Raises:
            TypeError: If the object type is not supported.
            FileNotFoundError: If the file is not found.
        """
        p = os.path.join(self.output_directory, path, filename)
        if not os.path.isfile(p):
            raise FileNotFoundError(f"File {p} not found.")
        _i = (
            torch.nn.Module,
            torch.optim.Optimizer,
            torch.optim.lr_scheduler.LRScheduler,
        )
        if not isinstance(obj, _i):
            raise TypeError(
                f"load_state of type {type(obj)} is not supported."
            )

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            state_dict = torch.load(p, map_location="cpu", weights_only=True)
        obj.load_state_dict(state_dict)

    def save_audobject(
        self,
        obj: audobject.Object,
        filename: str,
        path: str = "",
    ) -> None:
        """Save an audobject.Object to disk.

        Args:
            obj: Object to save.
            filename: Name of the file to save the object to.
            path: Subdirectory to save the object to. Defaults to "".

        Raises:
            TypeError: If the object type is not supported.
        """
        if not isinstance(obj, audobject.Object):
            raise TypeError(
                f"save_audobject of type {type(obj)} is not supported."
            )
        os.makedirs(os.path.join(self.output_directory, path), exist_ok=True)
        obj.to_yaml(os.path.join(self.output_directory, path, filename))

    def save_results_dict(
        self,
        results_dict: Dict[str, float],
        filename: str,
        path: str = "",
    ) -> None:
        """Save a results dictionary to disk.

        Args:
            results_dict: Dictionary of metric names and values to save.
            filename: Name of the file to save the results to.
            path: Subdirectory to save the results to. Defaults to "".
        """
        os.makedirs(os.path.join(self.output_directory, path), exist_ok=True)
        with open(
            os.path.join(self.output_directory, path, filename),
            "w",
            encoding="utf-8",
        ) as f:
            yaml.dump(results_dict, f)

    def save_results_df(
        self, results_df: pd.DataFrame, filename: str, path: str = ""
    ) -> None:
        """Save a results DataFrame to disk.

        Args:
            results_df: DataFrame to save.
            filename: Name of the file to save the results to.
            path: Subdirectory to save the results to. Defaults to "".
        """
        os.makedirs(os.path.join(self.output_directory, path), exist_ok=True)
        results_df.to_csv(
            os.path.join(self.output_directory, path, filename), index=False
        )

    def save_results_np(
        self, results_np: np.ndarray, filename: str, path: str = ""
    ) -> None:
        """Save a results numpy array to disk.

        Args:
            results_np: Numpy array to save.
            filename: Name of the file to save the results to.
            path: Subdirectory to save the results to. Defaults to "".
        """
        os.makedirs(os.path.join(self.output_directory, path), exist_ok=True)
        np.save(
            os.path.join(self.output_directory, path, filename), results_np
        )

    def save_best_results(
        self,
        metrics: pd.DataFrame,
        filename: str,
        metric_fns: List[AbstractMetric],
        tracking_metric_fn: AbstractMetric,
        path: str = "",
    ) -> None:
        """Save the best results to disk.

        Args:
            metrics: DataFrame of metrics to save.
            filename: Name of the file to save the best results to.
            metric_fns: List of metric functions to get the best results from.
            tracking_metric_fn: Tracking metric function to get the best
                iteration from.
            path: Subdirectory to save the best results to. Defaults to "".
        """
        best_metrics = {}
        for m in metrics:
            if "loss" in m:
                best_metrics[f"{m}_min"] = float(metrics[m].min())

        for m in metric_fns:
            best_metrics[m.name] = m.get_best(metrics[m.name])

        best_metrics["best_iteration"] = tracking_metric_fn.get_best_pos(
            metrics[tracking_metric_fn.name]
        )
        os.makedirs(os.path.join(self.output_directory, path), exist_ok=True)
        with open(
            os.path.join(self.output_directory, path, filename),
            "w",
            encoding="utf-8",
        ) as f:
            yaml.dump(best_metrics, f)



from typing import Union

import numpy as np
import torch


def _to_numpy(data: Union[torch.Tensor, np.ndarray]) -> np.ndarray:
    if isinstance(data, torch.Tensor):
        data = data.cpu().numpy()
    return data


def _to_tensor(data: Union[torch.Tensor, np.ndarray]) -> torch.Tensor:
    if isinstance(data, np.ndarray):
        data = torch.from_numpy(data)
    return data



import logging

import timm
import torch

from .abstract_model import AbstractModel
from .utils import ExtractLayerEmbeddings


class TimmModel(AbstractModel):
    def __init__(
        self,
        output_dim: int,
        timm_name: str,
        transfer=False,
        **kwargs,
    ) -> None:
        """Wrapper for timm models.

        Args:
            output_dim: Number of output classes.
            timm_name: Name of the model available in `timm.create_model`.
            transfer: Whether to load the model with pretrained weights.
                The final layer is replaced with a new layer with `output_dim`
                output features. Defaults to False.
            kwargs: Additional arguments to pass to the model constructor.
        """
        super().__init__(output_dim)
        self.timm_name = timm_name
        self.transfer = transfer
        for key, value in kwargs.items():
            setattr(self, key, value)

        logging.getLogger("timm").setLevel(logging.ERROR)
        self.model = timm.create_model(
            timm_name,
            pretrained=transfer,
            num_classes=output_dim,
            **kwargs,
        )
        self._embedding_extractor = ExtractLayerEmbeddings(self.model)

    def embeddings(self, x: torch.Tensor) -> torch.Tensor:
        return self._embedding_extractor(x)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)



from dataclasses import dataclass
import fnmatch
import os

import autrainer
from autrainer.core.constants import CONFIG_FOLDERS

from .abstract_script import AbstractScript, MockParser
from .command_line_error import CommandLineError
from .utils import catch_cli_errors


@dataclass
class ListArgs:
    directory: str
    local_only: bool
    global_only: bool
    pattern: str


class ListScript(AbstractScript):
    def __init__(self) -> None:
        super().__init__(
            "list",
            "List local and global configurations.",
            epilog="Example: autrainer list model -g -p EfficientNet-B*",
            dataclass=ListArgs,
        )

    def add_arguments(self) -> None:
        self.parser.add_argument(
            "directory",
            type=str,
            help="The directory to list configurations from. Choose from:"
            + "\n - ".join([""] + CONFIG_FOLDERS),
        )
        self.parser.add_argument(
            "-l",
            "--local-only",
            action="store_true",
            help="List local configurations only.",
        )
        self.parser.add_argument(
            "-g",
            "--global-only",
            action="store_true",
            help="List global configurations only.",
        )
        self.parser.add_argument(
            "-p",
            "--pattern",
            type=str,
            default="*",
            required=False,
            metavar="P",
            help="Glob pattern to filter configurations.",
        )

    def main(self, args: ListArgs) -> None:
        self._assert_config_directory(args.directory)
        self._list_configs(args)

    def _list_configs(self, args: ListArgs) -> None:
        if args.local_only == args.global_only:
            self._list_global_configs(args)
            self._list_local_configs(args)
        elif args.global_only:
            self._list_global_configs(args)
        else:
            self._list_local_configs(args)

    def _list_global_configs(self, args: ListArgs) -> None:
        path = os.path.join(
            os.path.dirname(autrainer.__path__[0]),
            "autrainer-configurations",
            args.directory,
        )
        self._print_configs(args.directory, path, "global", args.pattern)

    def _list_local_configs(self, args: ListArgs) -> None:
        path = os.path.join("conf", args.directory)
        if os.path.exists(path):
            self._print_configs(
                args.directory,
                os.path.join("conf", args.directory),
                "local",
                args.pattern,
            )
            return

        if args.local_only and not args.global_only:
            m = (
                f"Local conf directory '{args.directory}' does not exist.\n"
                f" Use 'autrainer create {args.directory}' "
                "to create a new project."
            )
            raise CommandLineError(self.parser, message=m, code=1)

    def _print_configs(
        self,
        directory: str,
        path: str,
        loc: str,
        pattern: str,
    ) -> None:
        files = [
            f.replace(".yaml", "")
            for f in sorted(os.listdir(path))
            if f.endswith(".yaml")
        ]
        files = fnmatch.filter(files, pattern)
        files = sorted(files, key=lambda x: (x.lower(), len(x)))
        if not files:
            print(f"No {loc} '{directory}' configurations found.")
            return
        print(f"{loc.capitalize()} '{directory}' configurations:")
        for f in files:
            print(f" - {f}")


@catch_cli_errors
def list_configs(
    directory: str,
    local_only: bool = False,
    global_only: bool = False,
    pattern: str = "*",
) -> None:
    """List local and global configurations.

    If called in a notebook, the function will not raise an error and print
    the error message instead.

    Args:
        directory: The directory to list configurations from. Choose from:
            :const:`~autrainer.core.constants.CONFIG_FOLDERS`.
        local_only: List local configurations only. Defaults to False.
        global_only: List global configurations only. Defaults to False.
        pattern: Glob pattern to filter configurations. Defaults to "*".

    Raises:
        CommandLineError: If the local configuration directory does not exist
            and local_only is True.
    """

    script = ListScript()
    script.parser = MockParser()
    script.main(ListArgs(directory, local_only, global_only, pattern))



from typing import Optional

from omegaconf import DictConfig, OmegaConf

import autrainer

from .abstract_script import AbstractScript, MockParser
from .utils import (
    add_hydra_args_to_sys,
    catch_cli_errors,
    run_hydra_cmd,
    running_in_notebook,
)


class GroupScript(AbstractScript):
    def __init__(self) -> None:
        super().__init__(
            "group",
            "Launch a manual grouping of multiple grid search results (Hydra).",
            extended_description=(
                "For more information on Hydra's command line line flags, see:\n"
                "https://hydra.cc/docs/advanced/hydra-command-line-flags/."
            ),
            epilog="Example: autrainer group -cn group.yaml",
            unknown_args=True,
        )

    def main(
        self,
        args: dict,
        config_name: str = "group",
        config_path: Optional[str] = None,
    ) -> None:
        @autrainer.main(config_name=config_name, config_path=config_path)
        def main(cfg: DictConfig) -> None:
            from autrainer.postprocessing import GroupGrid

            OmegaConf.set_struct(cfg, False)
            OmegaConf.resolve(cfg)
            ma = GroupGrid(
                results_dir=cfg.results_dir,
                groupings=cfg.groupings,
                max_runs=cfg.max_runs,
                plot_params=cfg.plotting,
            )

            ma.group_runs()

        main()


@catch_cli_errors
def group(
    override_kwargs: Optional[dict] = None,
    config_name: str = "group",
    config_path: Optional[str] = None,
) -> None:
    """Launch a manual grouping of multiple grid search results.

    Args:
        override_kwargs: Additional Hydra override arguments to pass to the
            train script.
        config_name: The name of the config (usually the file name without the
            .yaml extension). Defaults to "group".
        config_path: The config path, a directory where Hydra will search for
            config files. If config_path is None no directory is added to the
            search path. Defaults to None.
    """
    if running_in_notebook():
        run_hydra_cmd("group", override_kwargs, config_name, config_path)
    else:
        add_hydra_args_to_sys(override_kwargs, config_name, config_path)
        script = GroupScript()
        script.parser = MockParser()
        script.main({})



from .callback_manager import CallbackManager, CallbackSignature
from .outputs_tracker import OutputsTracker
from .training import ModularTaskTrainer


__all__ = [
    "CallbackManager",
    "CallbackSignature",
    "OutputsTracker",
    "ModularTaskTrainer",
]



from .abstract_model import AbstractModel
from .ast_model import ASTModel
from .cnn_10 import Cnn10
from .cnn_14 import Cnn14
from .end2you import AudioRNNModel
from .ffnn import FFNN
from .leaf import LEAFNet
from .sequential import SeqFFNN
from .tdnn import TDNNFFNN
from .timm_wrapper import TimmModel
from .torchvision_wrapper import TorchvisionModel
from .w2v2 import W2V2FFNN
from .whisper import WhisperFFNN


__all__ = [
    "AbstractModel",
    "ASTModel",
    "AudioRNNModel",
    "Cnn10",
    "Cnn14",
    "FFNN",
    "LEAFNet",
    "SeqFFNN",
    "TDNNFFNN",
    "TimmModel",
    "TorchvisionModel",
    "W2V2FFNN",
    "WhisperFFNN",
]



from .abstract_metric import (
    AbstractMetric,
    BaseAscendingMetric,
    BaseDescendingMetric,
)
from .classification import F1, UAR, Accuracy
from .multi_label_classification import (
    MLAccuracy,
    MLF1Macro,
    MLF1Micro,
    MLF1Weighted,
)
from .regression import CCC, MAE, MSE, PCC


__all__ = [
    "BaseAscendingMetric",
    "BaseDescendingMetric",
    "AbstractMetric",
    "F1",
    "UAR",
    "Accuracy",
    "MLAccuracy",
    "MLF1Macro",
    "MLF1Micro",
    "MLF1Weighted",
    "CCC",
    "MAE",
    "MSE",
    "PCC",
]



from functools import cached_property
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from omegaconf import DictConfig
import pandas as pd
import torch

from autrainer.transforms import SmartCompose

from .abstract_dataset import AbstractDataset
from .utils import (
    AbstractTargetTransform,
    LabelEncoder,
    MinMaxScaler,
    MultiLabelEncoder,
)


class ToyDatasetWrapper(torch.utils.data.Dataset):
    def __init__(
        self,
        df: pd.DataFrame,
        target_column: Union[list, str],
        feature_shape: Union[int, List[int]],
        dtype: str,
        generator: torch.Generator,
        transform: SmartCompose = None,
        target_transform: SmartCompose = None,
    ) -> None:
        self.df = df
        self.target_column = target_column
        self.feature_shape = feature_shape
        self.transform = transform
        self.target_transform = target_transform
        self.dtype = dtype
        self.generator = generator

    def __len__(self) -> int:
        return len(self.df)

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, int, int]:
        target = self.df.iloc[index][self.target_column]
        if isinstance(target, pd.Series):
            target = target.values
        if self.dtype == "float32":
            data = torch.rand(
                self.feature_shape,
                dtype=torch.float32,
                generator=self.generator,
            )
        else:
            data = torch.randint(
                0,
                255,
                self.feature_shape,
                dtype=torch.uint8,
                generator=self.generator,
            )
        if self.transform:
            img = self.transform(data, index=index)
        if self.target_transform:
            target = self.target_transform(target)
        return img, target, index


class ToyDataset(AbstractDataset):
    def __init__(
        self,
        task: str,
        size: int,
        num_targets: int,
        feature_shape: Union[int, List[int]],
        dev_split: float,
        test_split: float,
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        batch_size: int,
        dtype: str = "float32",
        inference_batch_size: Optional[int] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
    ) -> None:
        """Toy dataset for testing purposes.

        Args:
            task: Task of the dataset in ['classification', 'regression',
                'ml-classification'].
            size: Size of the dataset.
            num_targets: Number of targets.
            feature_shape: Shape of the features.
            dev_split: Proportion of the dataset to use for the development set.
            test_split: Proportion of the dataset to use for the test set.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            batch_size: Batch size.
            inference_batch_size: Inference batch size. If None, defaults to
                batch_size. Defaults to None.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
        """
        if task not in ["classification", "regression", "ml-classification"]:
            raise ValueError(
                f"Invalid task '{task}', must be in ['classification', "
                f"'regression', 'ml-classification']."
            )
        self.size = size
        self.num_targets = num_targets
        self.feature_shape = feature_shape
        self._assert_splits(dev_split, test_split)
        self.dev_split = dev_split
        self.test_split = test_split
        self._generator = torch.Generator().manual_seed(seed)
        if dtype not in ["float32", "uint8"]:
            raise ValueError(
                f"Invalid dtype='{dtype}', must be in ['float32', 'uint8']."
            )
        self.dtype = dtype
        super().__init__(
            task=task,
            path="/",
            features_subdir="",
            file_type="",
            index_column="index",
            target_column="",
            file_handler="autrainer.datasets.utils.IdentityFileHandler",
            seed=seed,
            metrics=metrics,
            tracking_metric=tracking_metric,
            batch_size=batch_size,
            inference_batch_size=inference_batch_size,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=None,
        )

    @staticmethod
    def _assert_splits(dev_split: int, test_split: float) -> None:
        if dev_split + test_split >= 1.0:
            raise ValueError(
                f"Sum of dev_split '{dev_split}' and test_split "
                f"'{test_split}' must be less than 1."
            )

    def _get_dataset_sizes(self) -> Tuple[int, int, int]:
        if self.size <= 0:
            raise ValueError(f"Size must be > 0, got '{self.size}'.")

        dev_size = int(self.size * self.dev_split)
        test_size = int(self.size * self.test_split)
        train_size = self.size - dev_size - test_size
        if any(size <= 0 for size in (train_size, dev_size, test_size)):
            raise ValueError(
                f"All subsets must be > 0, got train_size '{train_size}', "
                f"dev_size '{dev_size}', test_size '{test_size}'."
            )
        return train_size, dev_size, test_size

    def load_dataframes(
        self,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        rng = np.random.default_rng(self.seed)

        def _reset_index(df: pd.DataFrame) -> pd.DataFrame:
            return df.reset_index(drop=True).reset_index(
                names=[self.index_column]
            )

        if self.task == "ml-classification":
            self.target_column = [
                f"class_{i+1}" for i in range(self.num_targets)
            ]
            df = pd.DataFrame(
                {
                    **{
                        col: rng.integers(0, 2, size=self.size).astype(
                            np.float32
                        )
                        for col in self.target_column
                    }
                }
            )
        elif self.task == "classification":
            self.target_column = "target"
            df = pd.DataFrame(
                {
                    self.target_column: rng.integers(
                        0, self.num_targets, size=self.size
                    )
                }
            )
            df[self.target_column] = df[self.target_column].apply(
                lambda x: f"class_{x+1}"
            )
        else:
            # self.task == "regression"
            self.target_column = "target"
            df = pd.DataFrame({self.target_column: rng.random((self.size,))})

        train_size, dev_size, _ = self._get_dataset_sizes()
        return (
            _reset_index(df.iloc[:train_size]),
            _reset_index(df.iloc[train_size : train_size + dev_size]),
            _reset_index(df.iloc[train_size + dev_size :]),
        )

    def _init_dataset(
        self,
        df: pd.DataFrame,
        transform: SmartCompose,
    ) -> ToyDatasetWrapper:
        return ToyDatasetWrapper(
            df=df,
            target_column=self.target_column,
            feature_shape=self.feature_shape,
            dtype=self.dtype,
            generator=self._generator,
            transform=transform,
            target_transform=self.target_transform,
        )

    @cached_property
    def target_transform(self) -> AbstractTargetTransform:
        if self.task == "ml-classification":
            return MultiLabelEncoder(0.5, self.target_column)
        elif self.task == "classification":
            return LabelEncoder(
                self.df_train[self.target_column].unique().tolist()
            )
        return MinMaxScaler(
            minimum=self.df_train[self.target_column].min(),
            maximum=self.df_train[self.target_column].max(),
        )

    @cached_property
    def output_dim(self) -> int:
        if self.task in ["ml-classification", "classification"]:
            return self.num_targets
        return 1



from typing import Dict, List, Optional

import audobject
import torch

import autrainer

from .abstract_augmentation import AbstractAugmentation


class Sequential(AbstractAugmentation, audobject.Object):
    def __init__(
        self,
        sequence: List[Dict],
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
    ) -> None:
        """Create a fixed sequence of augmentations.

        The order of the augmentations in the list is not considered and is
        placed with respect to the order of the sequence augmentation itself.
        This means that the sequence of augmentations is applied in the order
        they are defined in the list and not disrupted by any other transform.

        Augmentations in the list must not have a collate function.

        Args:
            sequence: A list of (shorthand syntax) dictionaries defining the
                augmentation sequence.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
        """
        super().__init__(order, p, generator_seed)

        self.sequence = sequence
        self.augmentation_sequence = []

        for aug in self.sequence:
            if isinstance(aug, str):
                self.augmentation_sequence.append(
                    {aug: {"generator_seed": self.generator_seed}}
                )
            else:
                aug_name = next(iter(aug.keys()))
                if aug[aug_name].get("generator_seed") is None:
                    aug[aug_name]["generator_seed"] = self.generator_seed
                self.augmentation_sequence.append(aug)

        self.augmentation_sequence: List[AbstractAugmentation] = [
            autrainer.instantiate_shorthand(
                aug,
                instance_of=AbstractAugmentation,
            )
            for aug in self.sequence
        ]
        for aug in self.augmentation_sequence:
            if hasattr(aug, "get_collate_fn"):
                raise ValueError(
                    "Choice augmentations must not have a collate function."
                )

    def apply(self, x: torch.Tensor, index: int = None) -> torch.Tensor:
        """Apply all augmentations in sequence to the input tensor.

        Args:
            x: The input tensor.
            index: The index of the input tensor in the dataset.
                Defaults to None.

        Returns:
            The augmented tensor.
        """
        for aug in self.augmentation_sequence:
            x = aug(x, index)
        return x



import datetime
import os
import time

import yaml


class Timer:
    def __init__(self, output_directory: str, timer_type: str) -> None:
        """Timer to measure time of different parts of the training process.

        Args:
            output_directory: Directory to save the timer.yaml file to.
            timer_type: Name of the timer.
        """
        self.time_log = []
        self.start_time = None
        self.output_directory = output_directory
        self.timer_type = timer_type

    def start(self) -> None:
        """Start the timer."""
        self.start_time = time.time()

    def stop(self) -> None:
        """Stop the timer.

        Raises:
            ValueError: If the timer was not started.
        """
        if self.start_time is None:
            raise ValueError("Timer not yet started!")
        run_time = time.time() - self.start_time
        self.start_time = None
        self.time_log.append(run_time)

    def get_time_log(self) -> list:
        """Get the time log.

        Returns:
            List of times.
        """
        return self.time_log

    def get_mean_seconds(self) -> float:
        """Get the mean time in seconds.

        Returns:
            Mean time in seconds.
        """
        return sum(self.time_log) / len(self.time_log)

    def get_total_seconds(self) -> float:
        """Get the total time in seconds.

        Returns:
            Total time in seconds.
        """
        return sum(self.time_log)

    @classmethod
    def pretty_time(cls, seconds: float) -> str:
        """Convert seconds to a pretty string.

        Args:
            seconds: Time in seconds.

        Returns:
            Time in a pretty string format.
        """
        pretty = datetime.timedelta(seconds=int(seconds))
        return str(pretty)

    def save(self, path: str = "") -> dict:
        """Save and append the timer to timer.yaml.

        Args:
            path: Subdirectory to save the timer.yaml file to relative to the
                output directory. Defaults to "".

        Returns:
            Dictionary with mean and total time in seconds and pretty format.
        """
        os.makedirs(os.path.join(self.output_directory, path), exist_ok=True)
        out_path = os.path.join(self.output_directory, path, "timer.yaml")
        time_dict = {
            self.timer_type: {
                "mean": Timer.pretty_time(self.get_mean_seconds()),
                "mean_seconds": self.get_mean_seconds(),
                "total": Timer.pretty_time(self.get_total_seconds()),
                "total_seconds": self.get_total_seconds(),
            }
        }
        with open(out_path, "a") as f:
            yaml.dump(time_dict, f)



from abc import ABC
import logging
import os
import pickle
import shutil

import matplotlib.pyplot as plt
import seaborn as sns


try:
    from latex import escape

    LATEX_AVAILABLE = True
except ImportError:
    LATEX_AVAILABLE = False

    def escape(x):
        return x


class PlotBase(ABC):
    def __init__(
        self,
        output_directory: str,
        training_type: str,
        figsize: tuple,
        latex: bool,
        filetypes: list,
        pickle: bool,
        context: str,
        palette: str,
        replace_none: bool,
        add_titles: bool,
        add_xlabels: bool,
        add_ylabels: bool,
        rcParams: dict,
    ) -> None:
        """Base class for plotting.

        Args:
            output_directory: Output directory to save plots to.
            training_type: Type of training in ["Epoch", "Step"].
            figsize: Figure size in inches.
            latex: Whether to use LaTeX in plots. Requires the `latex` package.
                To install all necessary dependencies, run:
                `pip install autrainer[latex]`.
            filetypes: Filetypes to save plots as.
            pickle: Whether to save additional pickle files of the plots.
            context: Context for seaborn plots.
            palette: Color palette for seaborn plots.
            replace_none: Whether to replace "None" in labels with "~".
            add_titles: Whether to add titles to plots.
            add_xlabels: Whether to add x-labels to plots.
            add_ylabels: Whether to add y-labels to plots.
            rcParams: Additional Matplotlib rcParams to set.
        """
        self.output_directory = output_directory
        self.training_type = training_type.lower()
        self.figsize = figsize
        self.latex = latex
        self.filetypes = filetypes
        self.pickle = pickle
        self.context = context
        self.palette = palette
        self.replace_none = replace_none
        self.add_titles = add_titles
        self.add_xlabels = add_xlabels
        self.add_ylabels = add_ylabels
        sns.set_theme(
            context=self.context,
            style="whitegrid",
            palette=self.palette,
            rc=rcParams,
        )
        self._log = logging.getLogger(__name__)
        if not self.latex:
            return
        if not LATEX_AVAILABLE:
            self._log.warning(
                "LaTeX is not available as plot style. Install the required "
                "extras with 'pip install autrainer[latex]'."
            )
            self.latex = False
            return
        self._enable_latex()

    def save_plot(
        self,
        fig: plt.Figure,
        name: str,
        path: str = "",
        close: bool = True,
        tight_layout: bool = True,
    ) -> None:
        """Save a plot to the output directory.

        Args:
            fig: Matplotlib figure to save.
            name: Name of the plot.
            path: Path to save the plot to relative to the output directory.
            close: Whether to close the figure after saving.
            tight_layout: Whether to apply tight layout to the plot.
        """
        if self.replace_none:
            self._replace_none(fig)
        if self.latex:
            self._escape_latex(fig)
        base_path = os.path.join(self.output_directory, path)
        os.makedirs(base_path, exist_ok=True)
        if self.pickle:
            pkl_path = os.path.join(base_path, f"{name}.pkl")
            with open(pkl_path, "wb") as f:
                pickle.dump(fig, f)
        for filetype in self.filetypes:
            fpath = os.path.join(base_path, f"{name}.{filetype}")
            if tight_layout:
                plt.tight_layout()
            fig.savefig(fpath, dpi=300)
        if close:
            plt.close(fig)

    def _enable_latex(self) -> None:
        if shutil.which("latex") is None:
            self._log.warning(
                "LaTeX is not available on the system. "
                "Install a LaTeX distribution to use LaTeX in plots."
            )
            self.latex = False
            return
        plt.rcParams["text.usetex"] = True

    def _apply_to_labels(self, fig: plt.Figure, func: callable) -> None:
        if not fig.axes or fig.axes[0].legend_ is None:
            return
        for text in fig.axes[0].legend_.texts:
            text.set_text(func(text.get_text()))

    def _escape_latex(self, fig: plt.Figure) -> None:
        self._apply_to_labels(fig, escape)

    def _replace_none(self, fig: plt.Figure) -> None:
        def process_label(label):
            parts = label.split("_")
            parts = [part if part != "None" else "~" for part in parts]
            parts = "_".join(parts).replace("_~", "~").replace("~_", "~")
            return parts

        self._apply_to_labels(fig, process_label)

    def _add_label(
        self, ax: plt.Axes, x_label: str = None, y_label: str = None
    ) -> None:
        if x_label is not None:
            if self.add_xlabels:
                ax.set_xlabel(x_label)
            else:
                ax.set_xlabel("")
        if y_label is not None:
            if self.add_ylabels:
                ax.set_ylabel(y_label)
            else:
                ax.set_ylabel("")

    def _add_titles(self, ax: plt.Axes, title: str) -> None:
        if self.add_titles:
            ax.set_title(title)
        else:
            ax.set_title("")
